import pytest
from icupy import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 62:
    pytest.skip("ICU4C<62", allow_module_level=True)

import copy

from icupy import (
    ICUError, Locale, MeasureUnit, StringEnumeration, UErrorCode,
)
from icupy.number import NumberFormatter


def test_api():
    # [1]
    # static int32_t icu::MeasureUnit::getAvailable(
    #       const char *type,
    #       MeasureUnit *destArray,
    #       int32_t destCapacity,
    #       UErrorCode &errorCode
    # )
    dest_array = MeasureUnit.get_available("currency")
    assert isinstance(dest_array, list)
    assert len(dest_array) > 0
    assert all(isinstance(x, MeasureUnit) for x in dest_array)
    assert all(x.get_type() == "currency" for x in dest_array)

    # [2]
    # static int32_t icu::MeasureUnit::getAvailable(
    #       MeasureUnit *destArray,
    #       int32_t destCapacity,
    #       UErrorCode &errorCode
    # )
    dest_array = MeasureUnit.get_available()
    assert isinstance(dest_array, list)
    assert len(dest_array) > 0
    assert all(isinstance(x, MeasureUnit) for x in dest_array)

    # static StringEnumeration *icu::MeasureUnit::getAvailableTypes(
    #       UErrorCode &errorCode
    # )
    types = MeasureUnit.get_available_types()
    assert isinstance(types, StringEnumeration)
    assert len(types) > 0
    assert "currency" in types

    # const char *icu::MeasureUnit::getSubtype()
    unit = MeasureUnit.create_square_kilometer()
    result = unit.get_subtype()
    assert isinstance(result, str)
    assert result == "square-kilometer"

    # const char *icu::MeasureUnit::getType()
    result = unit.get_type()
    assert isinstance(result, str)
    assert result == "area"

    # UBool icu::MeasureUnit::operator!=(const UObject &other)
    unit1 = MeasureUnit.get_kilometer()
    unit2 = unit1.clone()
    unit3 = MeasureUnit.get_kilometer_per_hour()
    assert not (unit1 != unit2)
    assert unit1 != unit3
    assert unit2 != unit3

    # UBool icu::MeasureUnit::operator==(const UObject &other)
    assert unit1 == unit2
    assert not (unit1 == unit3)
    assert not (unit2 == unit3)


def test_clone():
    unit1 = MeasureUnit.get_kilometer()

    # MeasureUnit *icu::MeasureUnit::clone()
    unit2 = unit1.clone()
    assert isinstance(unit2, MeasureUnit)
    assert unit2 == unit1

    unit3 = copy.copy(unit1)
    assert unit3 == unit1

    unit4 = copy.deepcopy(unit1)
    assert unit4 == unit1


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_for_identifier():
    # static MeasureUnit icu::MeasureUnit::forIdentifier(
    #       StringPiece identifier,
    #       UErrorCode &status
    # )
    unit = MeasureUnit.for_identifier("furlong-per-nanosecond")
    assert isinstance(unit, MeasureUnit)

    # const char *icu::MeasureUnit::getIdentifier()
    result = unit.get_identifier()
    assert isinstance(result, str)
    assert result == "furlong-per-nanosecond"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_get_complexity():
    from icupy import UMeasureUnitComplexity

    # UMeasureUnitComplexity icu::MeasureUnit::getComplexity(
    #       UErrorCode &status
    # )
    unit = MeasureUnit.get_meter()
    assert (unit.get_complexity()
            == UMeasureUnitComplexity.UMEASURE_UNIT_SINGLE)

    unit = MeasureUnit.for_identifier("meter-per-second")
    assert (unit.get_complexity()
            == UMeasureUnitComplexity.UMEASURE_UNIT_COMPOUND)

    unit = MeasureUnit.for_identifier("hour-and-minute")
    assert (unit.get_complexity()
            == UMeasureUnitComplexity.UMEASURE_UNIT_MIXED)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_get_dimensionality():
    # int32_t icu::MeasureUnit::getDimensionality(UErrorCode &status)
    unit = MeasureUnit.create_cubic_meter()
    assert unit.get_dimensionality() == 3

    unit = MeasureUnit.create_meter()
    assert unit.get_dimensionality() == 1


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 69, reason="ICU4C<69")
def test_get_prefix():
    from icupy import UMeasurePrefix

    # UMeasurePrefix icu::MeasureUnit::getPrefix(UErrorCode &status)
    unit = MeasureUnit.get_kilometer()
    assert unit.get_prefix() == UMeasurePrefix.UMEASURE_PREFIX_KILO


def test_measure_unit_53():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_acre())
            .format_int(1)
            .to_string()) == "1 ac"

    assert (fmt.unit(MeasureUnit.create_arc_minute())
            .format_int(1)
            .to_string()) == "1 arcmin"

    assert (fmt.unit(MeasureUnit.create_arc_second())
            .format_int(1)
            .to_string()) == "1 arcsec"

    assert (fmt.unit(MeasureUnit.create_celsius())
            .format_int(1)
            .to_string()) == "1\xb0C"  # 1°C

    assert (fmt.unit(MeasureUnit.create_centimeter())
            .format_int(1)
            .to_string()) == "1 cm"

    assert (fmt.unit(MeasureUnit.create_cubic_kilometer())
            .format_int(1)
            .to_string()) == "1 km\xb3"  # 1 km³

    assert (fmt.unit(MeasureUnit.create_cubic_mile())
            .format_int(1)
            .to_string()) == "1 mi\xb3"  # 1 mi³

    assert (fmt.unit(MeasureUnit.create_day())
            .format_int(1)
            .to_string()) == "1 day"

    assert (fmt.unit(MeasureUnit.create_degree())
            .format_int(1)
            .to_string()) == "1 deg"

    assert (fmt.unit(MeasureUnit.create_fahrenheit())
            .format_int(1)
            .to_string()) == "1\xb0F"  # 1°F

    assert (fmt.unit(MeasureUnit.create_foot())
            .format_int(1)
            .to_string()) == "1 ft"

    assert (fmt.unit(MeasureUnit.create_gforce())
            .format_int(1)
            .to_string()) == "1 G"

    assert (fmt.unit(MeasureUnit.create_gram())
            .format_int(1)
            .to_string()) == "1 g"

    assert (fmt.unit(MeasureUnit.create_hectare())
            .format_int(1)
            .to_string()) == "1 ha"

    assert (fmt.unit(MeasureUnit.create_hectopascal())
            .format_int(1)
            .to_string()) == "1 hPa"

    assert (fmt.unit(MeasureUnit.create_horsepower())
            .format_int(1)
            .to_string()) == "1 hp"

    assert (fmt.unit(MeasureUnit.create_hour())
            .format_int(1)
            .to_string()) == "1 hr"

    assert (fmt.unit(MeasureUnit.create_inch())
            .format_int(1)
            .to_string()) == "1 in"

    assert (fmt.unit(MeasureUnit.create_inch_hg())
            .format_int(1)
            .to_string()) == "1 inHg"

    assert (fmt.unit(MeasureUnit.create_kilogram())
            .format_int(1)
            .to_string()) == "1 kg"

    assert (fmt.unit(MeasureUnit.create_kilometer())
            .format_int(1)
            .to_string()) == "1 km"

    assert (fmt.unit(MeasureUnit.create_kilometer_per_hour())
            .format_int(1)
            .to_string()) == "1 km/h"

    assert (fmt.unit(MeasureUnit.create_kilowatt())
            .format_int(1)
            .to_string()) == "1 kW"

    assert (fmt.unit(MeasureUnit.create_light_year())
            .format_int(1)
            .to_string()) == "1 ly"

    assert (fmt.unit(MeasureUnit.create_liter())
            .format_int(1)
            .to_string()) == "1 L"

    assert (fmt.unit(MeasureUnit.create_meter())
            .format_int(1)
            .to_string()) == "1 m"

    assert (fmt.unit(MeasureUnit.create_meter_per_second())
            .format_int(1)
            .to_string()) == "1 m/s"

    assert (fmt.unit(MeasureUnit.create_mile())
            .format_int(1)
            .to_string()) == "1 mi"

    assert (fmt.unit(MeasureUnit.create_mile_per_hour())
            .format_int(1)
            .to_string()) == "1 mph"

    assert (fmt.unit(MeasureUnit.create_millibar())
            .format_int(1)
            .to_string()) == "1 mbar"

    assert (fmt.unit(MeasureUnit.create_millimeter())
            .format_int(1)
            .to_string()) == "1 mm"

    assert (fmt.unit(MeasureUnit.create_millisecond())
            .format_int(1)
            .to_string()) == "1 ms"

    assert (fmt.unit(MeasureUnit.create_minute())
            .format_int(1)
            .to_string()) == "1 min"

    assert (fmt.unit(MeasureUnit.create_month())
            .format_int(1)
            .to_string()) == "1 mth"

    assert (fmt.unit(MeasureUnit.create_ounce())
            .format_int(1)
            .to_string()) == "1 oz"

    assert (fmt.unit(MeasureUnit.create_picometer())
            .format_int(1)
            .to_string()) == "1 pm"

    assert (fmt.unit(MeasureUnit.create_pound())
            .format_int(1)
            .to_string()) == "1 lb"

    assert (fmt.unit(MeasureUnit.create_second())
            .format_int(1)
            .to_string()) == "1 sec"

    assert (fmt.unit(MeasureUnit.create_square_foot())
            .format_int(1)
            .to_string()) == "1 sq ft"

    assert (fmt.unit(MeasureUnit.create_square_kilometer())
            .format_int(1)
            .to_string()) == "1 km\xb2"  # 1 km²

    assert (fmt.unit(MeasureUnit.create_square_meter())
            .format_int(1)
            .to_string()) == "1 m\xb2"  # 1 m²

    assert (fmt.unit(MeasureUnit.create_square_mile())
            .format_int(1)
            .to_string()) == "1 sq mi"

    assert (fmt.unit(MeasureUnit.create_watt())
            .format_int(1)
            .to_string()) == "1 W"

    assert (fmt.unit(MeasureUnit.create_week())
            .format_int(1)
            .to_string()) == "1 wk"

    assert (fmt.unit(MeasureUnit.create_yard())
            .format_int(1)
            .to_string()) == "1 yd"

    assert (fmt.unit(MeasureUnit.create_year())
            .format_int(1)
            .to_string()) == "1 yr"


def test_measure_unit_54():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_acre_foot())
            .format_int(1)
            .to_string()) == "1 ac ft"

    assert (fmt.unit(MeasureUnit.create_ampere())
            .format_int(1)
            .to_string()) == "1 A"

    assert (fmt.unit(MeasureUnit.create_astronomical_unit())
            .format_int(1)
            .to_string()) == "1 au"

    assert (fmt.unit(MeasureUnit.create_bit())
            .format_int(1)
            .to_string()) == "1 bit"

    assert (fmt.unit(MeasureUnit.create_bushel())
            .format_int(1)
            .to_string()) == "1 bu"

    assert (fmt.unit(MeasureUnit.create_byte())
            .format_int(1)
            .to_string()) == "1 byte"

    assert (fmt.unit(MeasureUnit.create_calorie())
            .format_int(1)
            .to_string()) == "1 cal"

    assert (fmt.unit(MeasureUnit.create_carat())
            .format_int(1)
            .to_string()) == "1 CD"

    assert (fmt.unit(MeasureUnit.create_centiliter())
            .format_int(1)
            .to_string()) == "1 cL"

    assert (fmt.unit(MeasureUnit.create_cubic_centimeter())
            .format_int(1)
            .to_string()) == "1 cm\xb3"  # 1 cm³

    assert (fmt.unit(MeasureUnit.create_cubic_foot())
            .format_int(1)
            .to_string()) == "1 ft\xb3"  # 1 ft³

    assert (fmt.unit(MeasureUnit.create_cubic_inch())
            .format_int(1)
            .to_string()) == "1 in\xb3"  # 1 in³

    assert (fmt.unit(MeasureUnit.create_cubic_meter())
            .format_int(1)
            .to_string()) == "1 m\xb3"  # 1 m³

    assert (fmt.unit(MeasureUnit.create_cubic_yard())
            .format_int(1)
            .to_string()) == "1 yd\xb3"  # 1 yd³

    assert (fmt.unit(MeasureUnit.create_cup())
            .format_int(1)
            .to_string()) == "1 c"

    assert (fmt.unit(MeasureUnit.create_deciliter())
            .format_int(1)
            .to_string()) == "1 dL"

    assert (fmt.unit(MeasureUnit.create_decimeter())
            .format_int(1)
            .to_string()) == "1 dm"

    assert (fmt.unit(MeasureUnit.create_fathom())
            .format_int(1)
            .to_string()) == "1 fth"

    assert (fmt.unit(MeasureUnit.create_fluid_ounce())
            .format_int(1)
            .to_string()) == "1 fl oz"

    assert (fmt.unit(MeasureUnit.create_foodcalorie())
            .format_int(1)
            .to_string()) == "1 Cal"

    assert (fmt.unit(MeasureUnit.create_furlong())
            .format_int(1)
            .to_string()) == "1 fur"

    assert (fmt.unit(MeasureUnit.create_gallon())
            .format_int(1)
            .to_string()) == "1 gal"

    assert (fmt.unit(MeasureUnit.create_gigabit())
            .format_int(1)
            .to_string()) == "1 Gb"

    assert (fmt.unit(MeasureUnit.create_gigabyte())
            .format_int(1)
            .to_string()) == "1 GB"

    assert (fmt.unit(MeasureUnit.create_gigahertz())
            .format_int(1)
            .to_string()) == "1 GHz"

    assert (fmt.unit(MeasureUnit.create_gigawatt())
            .format_int(1)
            .to_string()) == "1 GW"

    assert (fmt.unit(MeasureUnit.create_hectoliter())
            .format_int(1)
            .to_string()) == "1 hL"

    assert (fmt.unit(MeasureUnit.create_hertz())
            .format_int(1)
            .to_string()) == "1 Hz"

    assert (fmt.unit(MeasureUnit.create_joule())
            .format_int(1)
            .to_string()) == "1 J"

    assert (fmt.unit(MeasureUnit.create_karat())
            .format_int(1)
            .to_string()) == "1 kt"

    assert (fmt.unit(MeasureUnit.create_kelvin())
            .format_int(1)
            .to_string()) == "1 K"

    assert (fmt.unit(MeasureUnit.create_kilobit())
            .format_int(1)
            .to_string()) == "1 kb"

    assert (fmt.unit(MeasureUnit.create_kilobyte())
            .format_int(1)
            .to_string()) == "1 kB"

    assert (fmt.unit(MeasureUnit.create_kilocalorie())
            .format_int(1)
            .to_string()) == "1 kcal"

    assert (fmt.unit(MeasureUnit.create_kilohertz())
            .format_int(1)
            .to_string()) == "1 kHz"

    assert (fmt.unit(MeasureUnit.create_kilojoule())
            .format_int(1)
            .to_string()) == "1 kJ"

    assert (fmt.unit(MeasureUnit.create_kilowatt_hour())
            .format_int(1)
            .to_string()) == "1 kWh"

    assert (fmt.unit(MeasureUnit.create_liter_per_kilometer())
            .format_int(1)
            .to_string()) == "1 L/km"

    assert (fmt.unit(MeasureUnit.create_lux())
            .format_int(1)
            .to_string()) == "1 lx"

    assert (fmt.unit(MeasureUnit.create_megabit())
            .format_int(1)
            .to_string()) == "1 Mb"

    assert (fmt.unit(MeasureUnit.create_megabyte())
            .format_int(1)
            .to_string()) == "1 MB"

    assert (fmt.unit(MeasureUnit.create_megahertz())
            .format_int(1)
            .to_string()) == "1 MHz"

    assert (fmt.unit(MeasureUnit.create_megaliter())
            .format_int(1)
            .to_string()) == "1 ML"

    assert (fmt.unit(MeasureUnit.create_megawatt())
            .format_int(1)
            .to_string()) == "1 MW"

    assert (fmt.unit(MeasureUnit.create_meter_per_second_squared())
            .format_int(1)
            .to_string()) == "1 m/s\xb2"  # 1 m/s²

    assert (fmt.unit(MeasureUnit.create_metric_ton())
            .format_int(1)
            .to_string()) == "1 t"

    assert (fmt.unit(MeasureUnit.create_microgram())
            .format_int(1)
            .to_string()) in ("1 \xb5g", "1 \u03bcg")  # 1 µg

    assert (fmt.unit(MeasureUnit.create_micrometer())
            .format_int(1)
            .to_string()) in ("1 \xb5m", "1 \u03bcm")  # 1 µm

    assert (fmt.unit(MeasureUnit.create_microsecond())
            .format_int(1)
            .to_string()) in ("1 \xb5s", "1 \u03bcs")  # 1 µs

    assert (fmt.unit(MeasureUnit.create_mile_per_gallon())
            .format_int(1)
            .to_string()) == "1 mpg"

    assert (fmt.unit(MeasureUnit.create_milliampere())
            .format_int(1)
            .to_string()) == "1 mA"

    assert (fmt.unit(MeasureUnit.create_milligram())
            .format_int(1)
            .to_string()) == "1 mg"

    assert (fmt.unit(MeasureUnit.create_milliliter())
            .format_int(1)
            .to_string()) == "1 mL"

    assert (fmt.unit(MeasureUnit.create_millimeter_of_mercury())
            .format_int(1)
            .to_string()) == "1 mmHg"

    assert (fmt.unit(MeasureUnit.create_milliwatt())
            .format_int(1)
            .to_string()) == "1 mW"

    assert (fmt.unit(MeasureUnit.create_nanometer())
            .format_int(1)
            .to_string()) == "1 nm"

    assert (fmt.unit(MeasureUnit.create_nanosecond())
            .format_int(1)
            .to_string()) == "1 ns"

    assert (fmt.unit(MeasureUnit.create_nautical_mile())
            .format_int(1)
            .to_string()) == "1 nmi"

    assert (fmt.unit(MeasureUnit.create_ohm())
            .format_int(1)
            .to_string()) == "1 \u03a9"  # 1 Ω

    assert (fmt.unit(MeasureUnit.create_ounce_troy())
            .format_int(1)
            .to_string()) == "1 oz t"

    assert (fmt.unit(MeasureUnit.create_parsec())
            .format_int(1)
            .to_string()) == "1 pc"

    assert (fmt.unit(MeasureUnit.create_pint())
            .format_int(1)
            .to_string()) == "1 pt"

    assert (fmt.unit(MeasureUnit.create_pound_per_square_inch())
            .format_int(1)
            .to_string()) == "1 psi"

    assert (fmt.unit(MeasureUnit.create_quart())
            .format_int(1)
            .to_string()) == "1 qt"

    assert (fmt.unit(MeasureUnit.create_radian())
            .format_int(1)
            .to_string()) == "1 rad"

    assert (fmt.unit(MeasureUnit.create_square_centimeter())
            .format_int(1)
            .to_string()) == "1 cm\xb2"  # 1 cm²

    assert (fmt.unit(MeasureUnit.create_square_inch())
            .format_int(1)
            .to_string()) == "1 in\xb2"  # 1 in²

    assert (fmt.unit(MeasureUnit.create_square_yard())
            .format_int(1)
            .to_string()) == "1 yd\xb2"  # 1 yd²

    assert (fmt.unit(MeasureUnit.create_stone())
            .format_int(1)
            .to_string()) == "1 st"

    assert (fmt.unit(MeasureUnit.create_tablespoon())
            .format_int(1)
            .to_string()) == "1 tbsp"

    assert (fmt.unit(MeasureUnit.create_teaspoon())
            .format_int(1)
            .to_string()) == "1 tsp"

    assert (fmt.unit(MeasureUnit.create_terabit())
            .format_int(1)
            .to_string()) == "1 Tb"

    assert (fmt.unit(MeasureUnit.create_terabyte())
            .format_int(1)
            .to_string()) == "1 TB"

    assert (fmt.unit(MeasureUnit.create_ton())
            .format_int(1)
            .to_string()) == "1 tn"

    assert (fmt.unit(MeasureUnit.create_volt())
            .format_int(1)
            .to_string()) == "1 V"


def test_measure_unit_56():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_century())
            .format_int(1)
            .to_string()) == "1 c"

    assert (fmt.unit(MeasureUnit.create_cup_metric())
            .format_int(1)
            .to_string()) == "1 mc"

    assert (fmt.unit(MeasureUnit.create_generic_temperature())
            .format_int(1)
            .to_string()) == "1\xb0"  # 1°

    assert (fmt.unit(MeasureUnit.create_knot())
            .format_int(1)
            .to_string()) == "1 kn"

    assert (fmt.unit(MeasureUnit.create_liter_per_100_kilometers())
            .format_int(1)
            .to_string()) == "1 L/100 km"

    assert (fmt.unit(MeasureUnit.create_mile_scandinavian())
            .format_int(1)
            .to_string()) == "1 smi"

    assert (fmt.unit(MeasureUnit.create_pint_metric())
            .format_int(1)
            .to_string()) == "1 mpt"

    assert (fmt.unit(MeasureUnit.create_revolution_angle())
            .format_int(1)
            .to_string()) == "1 rev"


def test_measure_unit_57():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_gallon_imperial())
            .format_int(1)
            .to_string()) == "1 gal Imp."

    assert (fmt.unit(MeasureUnit.create_mile_per_gallon_imperial())
            .format_int(1)
            .to_string()) == "1 mpg Imp."

    try:
        assert (fmt.unit(MeasureUnit.create_milligram_per_deciliter())
                .format_int(1)
                .to_string()) == "1 mg/dL"
    except ICUError as ex:
        assert ex.args[0] == UErrorCode.U_MISSING_RESOURCE_ERROR  # ICU 69.1

    assert (fmt.unit(MeasureUnit.create_millimole_per_liter())
            .format_int(1)
            .to_string()) == "1 mmol/L"

    assert (fmt.unit(MeasureUnit.create_part_per_million())
            .format_int(1)
            .to_string()) == "1 ppm"


def test_measure_unit_59():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_point())
            .format_int(1)
            .to_string()) == "1 pt"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 63, reason="ICU4C<63")
def test_measure_unit_63():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_atmosphere())
            .format_int(1)
            .to_string()) == "1 atm"

    assert (fmt.unit(MeasureUnit.create_percent())
            .format_int(1)
            .to_string()) == "1%"

    assert (fmt.unit(MeasureUnit.create_permille())
            .format_int(1)
            .to_string()) == "1\u2030"  # 1‰

    assert (fmt.unit(MeasureUnit.create_petabyte())
            .format_int(1)
            .to_string()) == "1 PB"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_measure_unit_64():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_barrel())
            .format_int(1)
            .to_string()) == "1 bbl"

    assert (fmt.unit(MeasureUnit.create_british_thermal_unit())
            .format_int(1)
            .to_string()) == "1 Btu"

    assert (fmt.unit(MeasureUnit.create_dalton())
            .format_int(1)
            .to_string()) == "1 Da"

    assert (fmt.unit(MeasureUnit.create_day_person())
            .format_int(1)
            .to_string()) == "1 day"

    assert (fmt.unit(MeasureUnit.create_dunam())
            .format_int(1)
            .to_string()) == "1 dunam"

    assert (fmt.unit(MeasureUnit.create_earth_mass())
            .format_int(1)
            .to_string()) == "1 M\u2295"  # 1 M⊕

    assert (fmt.unit(MeasureUnit.create_electronvolt())
            .format_int(1)
            .to_string()) == "1 eV"

    assert (fmt.unit(MeasureUnit.create_fluid_ounce_imperial())
            .format_int(1)
            .to_string()) == "1 fl oz Imp."

    assert (fmt.unit(MeasureUnit.create_kilopascal())
            .format_int(1)
            .to_string()) == "1 kPa"

    assert (fmt.unit(MeasureUnit.create_megapascal())
            .format_int(1)
            .to_string()) == "1 MPa"

    assert (fmt.unit(MeasureUnit.create_mole())
            .format_int(1)
            .to_string()) == "1 mol"

    assert (fmt.unit(MeasureUnit.create_month_person())
            .format_int(1)
            .to_string()) == "1 mth"

    assert (fmt.unit(MeasureUnit.create_newton())
            .format_int(1)
            .to_string()) == "1 N"

    assert (fmt.unit(MeasureUnit.create_newton_meter())
            .format_int(1)
            .to_string()) == "1 N\u22c5m"  # 1 N⋅m

    assert (fmt.unit(MeasureUnit.create_permyriad())
            .format_int(1)
            .to_string()) == "1\u2031"  # 1‱

    assert (fmt.unit(MeasureUnit.create_pound_foot())
            .format_int(1)
            .to_string()) == "1 lbf\u22c5ft"  # 1 lb⋅ft

    assert (fmt.unit(MeasureUnit.create_pound_force())
            .format_int(1)
            .to_string()) == "1 lbf"

    assert (fmt.unit(MeasureUnit.create_solar_luminosity())
            .format_int(1)
            .to_string()) == "1 L\u2609"  # 1 L☉

    assert (fmt.unit(MeasureUnit.create_solar_mass())
            .format_int(1)
            .to_string()) == "1 M\u2609"  # 1 M☉

    assert (fmt.unit(MeasureUnit.create_solar_radius())
            .format_int(1)
            .to_string()) == "1 R\u2609"  # 1 R☉

    assert (fmt.unit(MeasureUnit.create_week_person())
            .format_int(1)
            .to_string()) == "1 wk"

    assert (fmt.unit(MeasureUnit.create_year_person())
            .format_int(1)
            .to_string()) == "1 yr"

    assert (fmt.unit(MeasureUnit.get_acre())
            .format_int(1)
            .to_string()) == "1 ac"

    assert (fmt.unit(MeasureUnit.get_acre_foot())
            .format_int(1)
            .to_string()) == "1 ac ft"

    assert (fmt.unit(MeasureUnit.get_ampere())
            .format_int(1)
            .to_string()) == "1 A"

    assert (fmt.unit(MeasureUnit.get_arc_minute())
            .format_int(1)
            .to_string()) == "1 arcmin"

    assert (fmt.unit(MeasureUnit.get_arc_second())
            .format_int(1)
            .to_string()) == "1 arcsec"

    assert (fmt.unit(MeasureUnit.get_astronomical_unit())
            .format_int(1)
            .to_string()) == "1 au"

    assert (fmt.unit(MeasureUnit.get_atmosphere())
            .format_int(1)
            .to_string()) == "1 atm"

    assert (fmt.unit(MeasureUnit.get_barrel())
            .format_int(1)
            .to_string()) == "1 bbl"

    assert (fmt.unit(MeasureUnit.get_bit())
            .format_int(1)
            .to_string()) == "1 bit"

    assert (fmt.unit(MeasureUnit.get_british_thermal_unit())
            .format_int(1)
            .to_string()) == "1 Btu"

    assert (fmt.unit(MeasureUnit.get_bushel())
            .format_int(1)
            .to_string()) == "1 bu"

    assert (fmt.unit(MeasureUnit.get_byte())
            .format_int(1)
            .to_string()) == "1 byte"

    assert (fmt.unit(MeasureUnit.get_calorie())
            .format_int(1)
            .to_string()) == "1 cal"

    assert (fmt.unit(MeasureUnit.get_carat())
            .format_int(1)
            .to_string()) == "1 CD"

    assert (fmt.unit(MeasureUnit.get_celsius())
            .format_int(1)
            .to_string()) == "1\xb0C"  # 1°C

    assert (fmt.unit(MeasureUnit.get_centiliter())
            .format_int(1)
            .to_string()) == "1 cL"

    assert (fmt.unit(MeasureUnit.get_centimeter())
            .format_int(1)
            .to_string()) == "1 cm"

    assert (fmt.unit(MeasureUnit.get_century())
            .format_int(1)
            .to_string()) == "1 c"

    assert (fmt.unit(MeasureUnit.get_cubic_centimeter())
            .format_int(1)
            .to_string()) == "1 cm\xb3"  # 1 cm³

    assert (fmt.unit(MeasureUnit.get_cubic_foot())
            .format_int(1)
            .to_string()) == "1 ft\xb3"  # 1 ft³

    assert (fmt.unit(MeasureUnit.get_cubic_inch())
            .format_int(1)
            .to_string()) == "1 in\xb3"  # 1 in³

    assert (fmt.unit(MeasureUnit.get_cubic_kilometer())
            .format_int(1)
            .to_string()) == "1 km\xb3"  # 1 km³

    assert (fmt.unit(MeasureUnit.get_cubic_meter())
            .format_int(1)
            .to_string()) == "1 m\xb3"  # 1 m³

    assert (fmt.unit(MeasureUnit.get_cubic_mile())
            .format_int(1)
            .to_string()) == "1 mi\xb3"  # 1 mi³

    assert (fmt.unit(MeasureUnit.get_cubic_yard())
            .format_int(1)
            .to_string()) == "1 yd\xb3"  # 1 yd³

    assert (fmt.unit(MeasureUnit.get_cup())
            .format_int(1)
            .to_string()) == "1 c"

    assert (fmt.unit(MeasureUnit.get_cup_metric())
            .format_int(1)
            .to_string()) == "1 mc"

    assert (fmt.unit(MeasureUnit.get_dalton())
            .format_int(1)
            .to_string()) == "1 Da"

    assert (fmt.unit(MeasureUnit.get_day())
            .format_int(1)
            .to_string()) == "1 day"

    assert (fmt.unit(MeasureUnit.get_day_person())
            .format_int(1)
            .to_string()) == "1 day"

    assert (fmt.unit(MeasureUnit.get_deciliter())
            .format_int(1)
            .to_string()) == "1 dL"

    assert (fmt.unit(MeasureUnit.get_decimeter())
            .format_int(1)
            .to_string()) == "1 dm"

    assert (fmt.unit(MeasureUnit.get_degree())
            .format_int(1)
            .to_string()) == "1 deg"

    assert (fmt.unit(MeasureUnit.get_dunam())
            .format_int(1)
            .to_string()) == "1 dunam"

    assert (fmt.unit(MeasureUnit.get_earth_mass())
            .format_int(1)
            .to_string()) == "1 M\u2295"  # 1 M⊕

    assert (fmt.unit(MeasureUnit.get_electronvolt())
            .format_int(1)
            .to_string()) == "1 eV"

    assert (fmt.unit(MeasureUnit.get_fahrenheit())
            .format_int(1)
            .to_string()) == "1\xb0F"  # 1°F

    assert (fmt.unit(MeasureUnit.get_fathom())
            .format_int(1)
            .to_string()) == "1 fth"

    assert (fmt.unit(MeasureUnit.get_fluid_ounce())
            .format_int(1)
            .to_string()) == "1 fl oz"

    assert (fmt.unit(MeasureUnit.get_fluid_ounce_imperial())
            .format_int(1)
            .to_string()) == "1 fl oz Imp."

    assert (fmt.unit(MeasureUnit.get_foodcalorie())
            .format_int(1)
            .to_string()) == "1 Cal"

    assert (fmt.unit(MeasureUnit.get_foot())
            .format_int(1)
            .to_string()) == "1 ft"

    assert (fmt.unit(MeasureUnit.get_furlong())
            .format_int(1)
            .to_string()) == "1 fur"

    assert (fmt.unit(MeasureUnit.get_gallon())
            .format_int(1)
            .to_string()) == "1 gal"

    assert (fmt.unit(MeasureUnit.get_gallon_imperial())
            .format_int(1)
            .to_string()) == "1 gal Imp."

    assert (fmt.unit(MeasureUnit.get_generic_temperature())
            .format_int(1)
            .to_string()) == "1\xb0"  # 1°

    assert (fmt.unit(MeasureUnit.get_gforce())
            .format_int(1)
            .to_string()) == "1 G"

    assert (fmt.unit(MeasureUnit.get_gigabit())
            .format_int(1)
            .to_string()) == "1 Gb"

    assert (fmt.unit(MeasureUnit.get_gigabyte())
            .format_int(1)
            .to_string()) == "1 GB"

    assert (fmt.unit(MeasureUnit.get_gigahertz())
            .format_int(1)
            .to_string()) == "1 GHz"

    assert (fmt.unit(MeasureUnit.get_gigawatt())
            .format_int(1)
            .to_string()) == "1 GW"

    assert (fmt.unit(MeasureUnit.get_gram())
            .format_int(1)
            .to_string()) == "1 g"

    assert (fmt.unit(MeasureUnit.get_hectare())
            .format_int(1)
            .to_string()) == "1 ha"

    assert (fmt.unit(MeasureUnit.get_hectoliter())
            .format_int(1)
            .to_string()) == "1 hL"

    assert (fmt.unit(MeasureUnit.get_hectopascal())
            .format_int(1)
            .to_string()) == "1 hPa"

    assert (fmt.unit(MeasureUnit.get_hertz())
            .format_int(1)
            .to_string()) == "1 Hz"

    assert (fmt.unit(MeasureUnit.get_horsepower())
            .format_int(1)
            .to_string()) == "1 hp"

    assert (fmt.unit(MeasureUnit.get_hour())
            .format_int(1)
            .to_string()) == "1 hr"

    assert (fmt.unit(MeasureUnit.get_inch())
            .format_int(1)
            .to_string()) == "1 in"

    assert (fmt.unit(MeasureUnit.get_inch_hg())
            .format_int(1)
            .to_string()) == "1 inHg"

    assert (fmt.unit(MeasureUnit.get_joule())
            .format_int(1)
            .to_string()) == "1 J"

    assert (fmt.unit(MeasureUnit.get_karat())
            .format_int(1)
            .to_string()) == "1 kt"

    assert (fmt.unit(MeasureUnit.get_kelvin())
            .format_int(1)
            .to_string()) == "1 K"

    assert (fmt.unit(MeasureUnit.get_kilobit())
            .format_int(1)
            .to_string()) == "1 kb"

    assert (fmt.unit(MeasureUnit.get_kilobyte())
            .format_int(1)
            .to_string()) == "1 kB"

    assert (fmt.unit(MeasureUnit.get_kilocalorie())
            .format_int(1)
            .to_string()) == "1 kcal"

    assert (fmt.unit(MeasureUnit.get_kilogram())
            .format_int(1)
            .to_string()) == "1 kg"

    assert (fmt.unit(MeasureUnit.get_kilohertz())
            .format_int(1)
            .to_string()) == "1 kHz"

    assert (fmt.unit(MeasureUnit.get_kilojoule())
            .format_int(1)
            .to_string()) == "1 kJ"

    assert (fmt.unit(MeasureUnit.get_kilometer())
            .format_int(1)
            .to_string()) == "1 km"

    assert (fmt.unit(MeasureUnit.get_kilometer_per_hour())
            .format_int(1)
            .to_string()) == "1 km/h"

    assert (fmt.unit(MeasureUnit.get_kilopascal())
            .format_int(1)
            .to_string()) == "1 kPa"

    assert (fmt.unit(MeasureUnit.get_kilowatt())
            .format_int(1)
            .to_string()) == "1 kW"

    assert (fmt.unit(MeasureUnit.get_kilowatt_hour())
            .format_int(1)
            .to_string()) == "1 kWh"

    assert (fmt.unit(MeasureUnit.get_knot())
            .format_int(1)
            .to_string()) == "1 kn"

    assert (fmt.unit(MeasureUnit.get_light_year())
            .format_int(1)
            .to_string()) == "1 ly"

    assert (fmt.unit(MeasureUnit.get_liter())
            .format_int(1)
            .to_string()) == "1 L"

    assert (fmt.unit(MeasureUnit.get_liter_per_100_kilometers())
            .format_int(1)
            .to_string()) == "1 L/100 km"

    assert (fmt.unit(MeasureUnit.get_liter_per_kilometer())
            .format_int(1)
            .to_string()) == "1 L/km"

    assert (fmt.unit(MeasureUnit.get_lux())
            .format_int(1)
            .to_string()) == "1 lx"

    assert (fmt.unit(MeasureUnit.get_megabit())
            .format_int(1)
            .to_string()) == "1 Mb"

    assert (fmt.unit(MeasureUnit.get_megabyte())
            .format_int(1)
            .to_string()) == "1 MB"

    assert (fmt.unit(MeasureUnit.get_megahertz())
            .format_int(1)
            .to_string()) == "1 MHz"

    assert (fmt.unit(MeasureUnit.get_megaliter())
            .format_int(1)
            .to_string()) == "1 ML"

    assert (fmt.unit(MeasureUnit.get_megapascal())
            .format_int(1)
            .to_string()) == "1 MPa"

    assert (fmt.unit(MeasureUnit.get_megawatt())
            .format_int(1)
            .to_string()) == "1 MW"

    assert (fmt.unit(MeasureUnit.get_meter())
            .format_int(1)
            .to_string()) == "1 m"

    assert (fmt.unit(MeasureUnit.get_meter_per_second())
            .format_int(1)
            .to_string()) == "1 m/s"

    assert (fmt.unit(MeasureUnit.get_meter_per_second_squared())
            .format_int(1)
            .to_string()) == "1 m/s\xb2"  # 1 m/s²

    assert (fmt.unit(MeasureUnit.get_metric_ton())
            .format_int(1)
            .to_string()) == "1 t"

    assert (fmt.unit(MeasureUnit.get_microgram())
            .format_int(1)
            .to_string()) in ("1 \xb5g", "1 \u03bcg")  # 1 µg

    assert (fmt.unit(MeasureUnit.get_micrometer())
            .format_int(1)
            .to_string()) in ("1 \xb5m", "1 \u03bcm")  # 1 µm

    assert (fmt.unit(MeasureUnit.get_microsecond())
            .format_int(1)
            .to_string()) in ("1 \xb5s", "1 \u03bcs")  # 1 µs

    assert (fmt.unit(MeasureUnit.get_mile())
            .format_int(1)
            .to_string()) == "1 mi"

    assert (fmt.unit(MeasureUnit.get_mile_per_gallon())
            .format_int(1)
            .to_string()) == "1 mpg"

    assert (fmt.unit(MeasureUnit.get_mile_per_gallon_imperial())
            .format_int(1)
            .to_string()) == "1 mpg Imp."

    assert (fmt.unit(MeasureUnit.get_mile_per_hour())
            .format_int(1)
            .to_string()) == "1 mph"

    assert (fmt.unit(MeasureUnit.get_mile_scandinavian())
            .format_int(1)
            .to_string()) == "1 smi"

    assert (fmt.unit(MeasureUnit.get_milliampere())
            .format_int(1)
            .to_string()) == "1 mA"

    assert (fmt.unit(MeasureUnit.get_millibar())
            .format_int(1)
            .to_string()) == "1 mbar"

    assert (fmt.unit(MeasureUnit.get_milligram())
            .format_int(1)
            .to_string()) == "1 mg"

    try:
        assert (fmt.unit(MeasureUnit.get_milligram_per_deciliter())
                .format_int(1)
                .to_string()) == "1 mg/dL"
    except ICUError as ex:
        assert ex.args[0] == UErrorCode.U_MISSING_RESOURCE_ERROR  # ICU 69.1

    assert (fmt.unit(MeasureUnit.get_milliliter())
            .format_int(1)
            .to_string()) == "1 mL"

    assert (fmt.unit(MeasureUnit.get_millimeter())
            .format_int(1)
            .to_string()) == "1 mm"

    assert (fmt.unit(MeasureUnit.get_millimeter_of_mercury())
            .format_int(1)
            .to_string()) == "1 mmHg"

    assert (fmt.unit(MeasureUnit.get_millimole_per_liter())
            .format_int(1)
            .to_string()) == "1 mmol/L"

    assert (fmt.unit(MeasureUnit.get_millisecond())
            .format_int(1)
            .to_string()) == "1 ms"

    assert (fmt.unit(MeasureUnit.get_milliwatt())
            .format_int(1)
            .to_string()) == "1 mW"

    assert (fmt.unit(MeasureUnit.get_minute())
            .format_int(1)
            .to_string()) == "1 min"

    assert (fmt.unit(MeasureUnit.get_mole())
            .format_int(1)
            .to_string()) == "1 mol"

    assert (fmt.unit(MeasureUnit.get_month())
            .format_int(1)
            .to_string()) == "1 mth"

    assert (fmt.unit(MeasureUnit.get_month_person())
            .format_int(1)
            .to_string()) == "1 mth"

    assert (fmt.unit(MeasureUnit.get_nanometer())
            .format_int(1)
            .to_string()) == "1 nm"

    assert (fmt.unit(MeasureUnit.get_nanosecond())
            .format_int(1)
            .to_string()) == "1 ns"

    assert (fmt.unit(MeasureUnit.get_nautical_mile())
            .format_int(1)
            .to_string()) == "1 nmi"

    assert (fmt.unit(MeasureUnit.get_newton())
            .format_int(1)
            .to_string()) == "1 N"

    assert (fmt.unit(MeasureUnit.get_newton_meter())
            .format_int(1)
            .to_string()) == "1 N\u22c5m"  # 1 N⋅m

    assert (fmt.unit(MeasureUnit.get_ohm())
            .format_int(1)
            .to_string()) == "1 \u03a9"  # 1 Ω

    assert (fmt.unit(MeasureUnit.get_ounce())
            .format_int(1)
            .to_string()) == "1 oz"

    assert (fmt.unit(MeasureUnit.get_ounce_troy())
            .format_int(1)
            .to_string()) == "1 oz t"

    assert (fmt.unit(MeasureUnit.get_parsec())
            .format_int(1)
            .to_string()) == "1 pc"

    assert (fmt.unit(MeasureUnit.get_part_per_million())
            .format_int(1)
            .to_string()) == "1 ppm"

    assert (fmt.unit(MeasureUnit.get_percent())
            .format_int(1)
            .to_string()) == "1%"

    assert (fmt.unit(MeasureUnit.get_permille())
            .format_int(1)
            .to_string()) == "1\u2030"  # 1‰

    assert (fmt.unit(MeasureUnit.get_permyriad())
            .format_int(1)
            .to_string()) == "1\u2031"  # 1‱

    assert (fmt.unit(MeasureUnit.get_petabyte())
            .format_int(1)
            .to_string()) == "1 PB"

    assert (fmt.unit(MeasureUnit.get_picometer())
            .format_int(1)
            .to_string()) == "1 pm"

    assert (fmt.unit(MeasureUnit.get_pint())
            .format_int(1)
            .to_string()) == "1 pt"

    assert (fmt.unit(MeasureUnit.get_pint_metric())
            .format_int(1)
            .to_string()) == "1 mpt"

    assert (fmt.unit(MeasureUnit.get_point())
            .format_int(1)
            .to_string()) == "1 pt"

    assert (fmt.unit(MeasureUnit.get_pound())
            .format_int(1)
            .to_string()) == "1 lb"

    assert (fmt.unit(MeasureUnit.get_pound_foot())
            .format_int(1)
            .to_string()) == "1 lbf\u22c5ft"  # 1 lb⋅ft

    assert (fmt.unit(MeasureUnit.get_pound_force())
            .format_int(1)
            .to_string()) == "1 lbf"

    assert (fmt.unit(MeasureUnit.get_pound_per_square_inch())
            .format_int(1)
            .to_string()) == "1 psi"

    assert (fmt.unit(MeasureUnit.get_quart())
            .format_int(1)
            .to_string()) == "1 qt"

    assert (fmt.unit(MeasureUnit.get_radian())
            .format_int(1)
            .to_string()) == "1 rad"

    assert (fmt.unit(MeasureUnit.get_revolution_angle())
            .format_int(1)
            .to_string()) == "1 rev"

    assert (fmt.unit(MeasureUnit.get_second())
            .format_int(1)
            .to_string()) == "1 sec"

    assert (fmt.unit(MeasureUnit.get_solar_luminosity())
            .format_int(1)
            .to_string()) == "1 L\u2609"  # 1 L☉

    assert (fmt.unit(MeasureUnit.get_solar_mass())
            .format_int(1)
            .to_string()) == "1 M\u2609"  # 1 M☉

    assert (fmt.unit(MeasureUnit.get_solar_radius())
            .format_int(1)
            .to_string()) == "1 R\u2609"  # 1 R☉

    assert (fmt.unit(MeasureUnit.get_square_centimeter())
            .format_int(1)
            .to_string()) == "1 cm\xb2"  # 1 cm²

    assert (fmt.unit(MeasureUnit.get_square_foot())
            .format_int(1)
            .to_string()) == "1 sq ft"

    assert (fmt.unit(MeasureUnit.get_square_inch())
            .format_int(1)
            .to_string()) == "1 in\xb2"  # 1 in²

    assert (fmt.unit(MeasureUnit.get_square_kilometer())
            .format_int(1)
            .to_string()) == "1 km\xb2"  # 1 km²

    assert (fmt.unit(MeasureUnit.get_square_meter())
            .format_int(1)
            .to_string()) == "1 m\xb2"  # 1 m²

    assert (fmt.unit(MeasureUnit.get_square_mile())
            .format_int(1)
            .to_string()) == "1 sq mi"

    assert (fmt.unit(MeasureUnit.get_square_yard())
            .format_int(1)
            .to_string()) == "1 yd\xb2"  # 1 yd²

    assert (fmt.unit(MeasureUnit.get_stone())
            .format_int(1)
            .to_string()) == "1 st"

    assert (fmt.unit(MeasureUnit.get_tablespoon())
            .format_int(1)
            .to_string()) == "1 tbsp"

    assert (fmt.unit(MeasureUnit.get_teaspoon())
            .format_int(1)
            .to_string()) == "1 tsp"

    assert (fmt.unit(MeasureUnit.get_terabit())
            .format_int(1)
            .to_string()) == "1 Tb"

    assert (fmt.unit(MeasureUnit.get_terabyte())
            .format_int(1)
            .to_string()) == "1 TB"

    assert (fmt.unit(MeasureUnit.get_ton())
            .format_int(1)
            .to_string()) == "1 tn"

    assert (fmt.unit(MeasureUnit.get_volt())
            .format_int(1)
            .to_string()) == "1 V"

    assert (fmt.unit(MeasureUnit.get_watt())
            .format_int(1)
            .to_string()) == "1 W"

    assert (fmt.unit(MeasureUnit.get_week())
            .format_int(1)
            .to_string()) == "1 wk"

    assert (fmt.unit(MeasureUnit.get_week_person())
            .format_int(1)
            .to_string()) == "1 wk"

    assert (fmt.unit(MeasureUnit.get_yard())
            .format_int(1)
            .to_string()) == "1 yd"

    assert (fmt.unit(MeasureUnit.get_year())
            .format_int(1)
            .to_string()) == "1 yr"

    assert (fmt.unit(MeasureUnit.get_year_person())
            .format_int(1)
            .to_string()) == "1 yr"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 65, reason="ICU4C<65")
def test_measure_unit_65():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_bar())
            .format_int(1)
            .to_string()) == "1 bar"

    assert (fmt.unit(MeasureUnit.create_decade())
            .format_int(1)
            .to_string()) == "1 dec"

    assert (fmt.unit(MeasureUnit.create_dot_per_centimeter())
            .format_int(1)
            .to_string()) == "1 dpcm"

    assert (fmt.unit(MeasureUnit.create_dot_per_inch())
            .format_int(1)
            .to_string()) == "1 dpi"

    assert (fmt.unit(MeasureUnit.create_em())
            .format_int(1)
            .to_string()) == "1 em"

    assert (fmt.unit(MeasureUnit.create_megapixel())
            .format_int(1)
            .to_string()) == "1 MP"

    assert (fmt.unit(MeasureUnit.create_pascal())
            .format_int(1)
            .to_string()) == "1 Pa"

    assert (fmt.unit(MeasureUnit.create_pixel())
            .format_int(1)
            .to_string()) == "1 px"

    assert (fmt.unit(MeasureUnit.create_pixel_per_centimeter())
            .format_int(1)
            .to_string()) == "1 ppcm"

    assert (fmt.unit(MeasureUnit.create_pixel_per_inch())
            .format_int(1)
            .to_string()) == "1 ppi"

    assert (fmt.unit(MeasureUnit.create_therm_us())
            .format_int(1)
            .to_string()) == "1 US therm"

    assert (fmt.unit(MeasureUnit.get_bar())
            .format_int(1)
            .to_string()) == "1 bar"

    assert (fmt.unit(MeasureUnit.get_decade())
            .format_int(1)
            .to_string()) == "1 dec"

    assert (fmt.unit(MeasureUnit.get_dot_per_centimeter())
            .format_int(1)
            .to_string()) == "1 dpcm"

    assert (fmt.unit(MeasureUnit.get_dot_per_inch())
            .format_int(1)
            .to_string()) == "1 dpi"

    assert (fmt.unit(MeasureUnit.get_em())
            .format_int(1)
            .to_string()) == "1 em"

    assert (fmt.unit(MeasureUnit.get_megapixel())
            .format_int(1)
            .to_string()) == "1 MP"

    assert (fmt.unit(MeasureUnit.get_pascal())
            .format_int(1)
            .to_string()) == "1 Pa"

    assert (fmt.unit(MeasureUnit.get_pixel())
            .format_int(1)
            .to_string()) == "1 px"

    assert (fmt.unit(MeasureUnit.get_pixel_per_centimeter())
            .format_int(1)
            .to_string()) == "1 ppcm"

    assert (fmt.unit(MeasureUnit.get_pixel_per_inch())
            .format_int(1)
            .to_string()) == "1 ppi"

    assert (fmt.unit(MeasureUnit.get_therm_us())
            .format_int(1)
            .to_string()) == "1 US therm"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 68, reason="ICU4C<68")
def test_measure_unit_68():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_candela())
            .format_int(1)
            .to_string()) == "1 cd"

    assert (fmt.unit(MeasureUnit.create_dessert_spoon())
            .format_int(1)
            .to_string()) == "1 dstspn"

    assert (fmt.unit(MeasureUnit.create_dessert_spoon_imperial())
            .format_int(1)
            .to_string()) == "1 dstspn Imp"

    assert (fmt.unit(MeasureUnit.create_dot())
            .format_int(1)
            .to_string()) in ("1 dot",  # ICU4C<69
                              "1 px")  # ICU4C>=69

    assert (fmt.unit(MeasureUnit.create_dram())
            .format_int(1)
            .to_string()) == "1 dram fl"

    assert (fmt.unit(MeasureUnit.create_drop())
            .format_int(1)
            .to_string()) == "1 drop"

    assert (fmt.unit(MeasureUnit.create_earth_radius())
            .format_int(1)
            .to_string()) == "1 R\u2295"  # 1 R⊕

    assert (fmt.unit(MeasureUnit.create_grain())
            .format_int(1)
            .to_string()) == "1 grain"

    assert (fmt.unit(MeasureUnit.create_jigger())
            .format_int(1)
            .to_string()) == "1 jigger"

    assert (fmt.unit(MeasureUnit.create_lumen())
            .format_int(1)
            .to_string()) == "1 lm"

    assert (fmt.unit(MeasureUnit.create_pinch())
            .format_int(1)
            .to_string()) == "1 pinch"

    assert (fmt.unit(MeasureUnit.create_quart_imperial())
            .format_int(1)
            .to_string()) == "1 qt Imp."

    assert (fmt.unit(MeasureUnit.get_candela())
            .format_int(1)
            .to_string()) == "1 cd"

    assert (fmt.unit(MeasureUnit.get_dessert_spoon())
            .format_int(1)
            .to_string()) == "1 dstspn"

    assert (fmt.unit(MeasureUnit.get_dessert_spoon_imperial())
            .format_int(1)
            .to_string()) == "1 dstspn Imp"

    assert (fmt.unit(MeasureUnit.get_dot())
            .format_int(1)
            .to_string()) in ("1 dot",  # ICU4C<69
                              "1 px")  # ICU4C>=69

    assert (fmt.unit(MeasureUnit.get_dram())
            .format_int(1)
            .to_string()) == "1 dram fl"

    assert (fmt.unit(MeasureUnit.get_drop())
            .format_int(1)
            .to_string()) == "1 drop"

    assert (fmt.unit(MeasureUnit.get_earth_radius())
            .format_int(1)
            .to_string()) == "1 R\u2295"  # 1 R⊕

    assert (fmt.unit(MeasureUnit.get_grain())
            .format_int(1)
            .to_string()) == "1 grain"

    assert (fmt.unit(MeasureUnit.get_jigger())
            .format_int(1)
            .to_string()) == "1 jigger"

    assert (fmt.unit(MeasureUnit.get_lumen())
            .format_int(1)
            .to_string()) == "1 lm"

    assert (fmt.unit(MeasureUnit.get_pinch())
            .format_int(1)
            .to_string()) == "1 pinch"

    assert (fmt.unit(MeasureUnit.get_quart_imperial())
            .format_int(1)
            .to_string()) == "1 qt Imp."


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 69, reason="ICU4C<69")
def test_measure_unit_69():
    fmt = NumberFormatter.with_locale(Locale.get_us())

    assert (fmt.unit(MeasureUnit.create_milligram_of_glucose_per_deciliter())
            .format_int(1)
            .to_string()) == "1 mg/dL"

    assert (fmt.unit(MeasureUnit.get_milligram_of_glucose_per_deciliter())
            .format_int(1)
            .to_string()) == "1 mg/dL"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_product():
    # MeasureUnit icu::MeasureUnit::product(
    #       const MeasureUnit &other,
    #       UErrorCode &status
    # )
    unit = MeasureUnit.for_identifier("kilowatt")
    unit2 = unit.product(MeasureUnit.for_identifier("hour-per-day"))
    assert isinstance(unit2, MeasureUnit)
    assert unit2.get_identifier() in ("hour-kilowatt-per-day",  # ICU4C<69
                                      "kilowatt-hour-per-day")  # ICU4C>=69


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_reciprocal():
    # MeasureUnit icu::MeasureUnit::reciprocal(UErrorCode &status)
    unit = MeasureUnit.for_identifier("meter-per-second")
    unit2 = unit.reciprocal()
    assert isinstance(unit2, MeasureUnit)
    assert unit2.get_identifier() == "second-per-meter"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 68, reason="ICU4C<68")
def test_split_to_single_units():
    unit = MeasureUnit.create_newton_meter()

    # std::pair<LocalArray<MeasureUnit>, int32_t>
    # icu::MeasureUnit::splitToSingleUnits(UErrorCode &status)
    units = unit.split_to_single_units()
    assert isinstance(units, list)
    assert len(units) == 2
    assert all(isinstance(x, MeasureUnit) for x in units)
    assert units[0].get_identifier() == "newton"
    assert units[1].get_identifier() == "meter"

    unit = MeasureUnit.create_kilometer()
    units = unit.split_to_single_units()
    assert isinstance(units, list)
    assert len(units) == 1
    assert all(isinstance(x, MeasureUnit) for x in units)
    assert units[0].get_identifier() == "kilometer"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_with_dimensionality():
    unit = MeasureUnit.create_kilometer()

    # MeasureUnit icu::MeasureUnit::withDimensionality(
    #       int32_t dimensionality,
    #       UErrorCode &status
    # )
    unit2 = unit.with_dimensionality(2)
    assert isinstance(unit2, MeasureUnit)
    assert unit2.get_identifier() == "square-kilometer"

    unit3 = unit.with_dimensionality(3)
    assert isinstance(unit3, MeasureUnit)
    assert unit3.get_identifier() == "cubic-kilometer"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 69, reason="ICU4C<69")
def test_with_prefix():
    from icupy import UMeasurePrefix

    unit = MeasureUnit.create_byte()

    # MeasureUnit icu::MeasureUnit::withPrefix(
    #       UMeasurePrefix prefix,
    #       UErrorCode &status
    # )
    unit2 = unit.with_prefix(UMeasurePrefix.UMEASURE_PREFIX_KILO)
    assert isinstance(unit2, MeasureUnit)
    assert unit2.get_identifier() == "kilobyte"



import copy
from functools import cmp_to_key

import pytest
from icupy import (
    CollationElementIterator, CollationKey, Collator, Locale,
    RuleBasedCollator, StringCharacterIterator, StringEnumeration,
    UColAttribute, UColAttributeValue, UColBoundMode, UColReorderCode,
    UColRuleOption, UCollationResult, UScriptCode,
    U_ICU_VERSION_MAJOR_NUM,
    UnicodeSet, UnicodeString,
)


def test_clone():
    test1 = Collator.create_instance(Locale.get_japanese())
    rules1 = test1.get_rules()

    # RuleBasedCollator *icu::RuleBasedCollator::clone()
    test2 = test1.clone()
    assert isinstance(test2, RuleBasedCollator)
    rules2 = test2.get_rules()
    assert rules2 == rules1

    test3 = copy.copy(test1)
    rules3 = test3.get_rules()
    assert rules3 == rules1

    test4 = copy.deepcopy(test1)
    rules4 = test4.get_rules()
    assert rules4 == rules1


def test_compare():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # [2]
    # UCollationResult icu::Collator::compare(
    #       const char16_t *source,
    #       int32_t sourceLength,
    #       const char16_t *target,
    #       int32_t targetLength,
    #       UErrorCode &status
    # )
    assert coll.compare("ABC", -1, "abc", -1) == UCollationResult.UCOL_GREATER

    # [5]
    # UCollationResult icu::Collator::compare(
    #       const UnicodeString &source,
    #       const UnicodeString &target,
    #       int32_t length,
    #       UErrorCode &status
    # )
    assert coll.compare(source, target, 3) == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", target, 3) == UCollationResult.UCOL_GREATER
    assert coll.compare(source, "abc", 3) == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", "abc", 3) == UCollationResult.UCOL_GREATER

    # [6]
    # UCollationResult icu::Collator::compare(
    #       const UnicodeString &source,
    #       const UnicodeString &target,
    #       UErrorCode &status
    # )
    assert coll.compare(source, target) == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", target) == UCollationResult.UCOL_GREATER
    assert coll.compare(source, "abc") == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", "abc") == UCollationResult.UCOL_GREATER

    coll.set_attribute(UColAttribute.UCOL_STRENGTH,
                       UColAttributeValue.UCOL_PRIMARY)
    assert coll.compare("ABC", -1, "abc", -1) == UCollationResult.UCOL_EQUAL
    assert coll.compare(source, target, 3) == UCollationResult.UCOL_EQUAL
    assert coll.compare(source, target) == UCollationResult.UCOL_EQUAL


def test_create_collation_element_iterator():
    coll = Collator.create_instance(Locale("es"))

    # [1]
    # CollationElementIterator *
    # icu::RuleBasedCollator::createCollationElementIterator(
    #       const CharacterIterator &source
    # )
    source = StringCharacterIterator(UnicodeString("cha"))
    it1 = coll.create_collation_element_iterator(source)
    assert isinstance(it1, CollationElementIterator)

    order1 = it1.next()
    assert order1 != CollationElementIterator.NULLORDER
    order2 = it1.next()
    assert order2 != CollationElementIterator.NULLORDER
    order3 = it1.next()
    assert order3 != CollationElementIterator.NULLORDER
    assert it1.next() == CollationElementIterator.NULLORDER

    it1.set_offset(3)
    assert it1.get_offset() == 3
    assert it1.previous() == order3
    assert it1.previous() == order2
    assert it1.previous() == order1
    assert it1.previous() == CollationElementIterator.NULLORDER

    assert list(it1) == [order1, order2, order3]

    # [2]
    # CollationElementIterator *
    # icu::RuleBasedCollator::createCollationElementIterator(
    #       const UnicodeString &source
    # )
    source = UnicodeString("cha")
    it2 = coll.create_collation_element_iterator(source)
    assert isinstance(it2, CollationElementIterator)

    order1 = it2.next()
    assert order1 != CollationElementIterator.NULLORDER
    order2 = it2.next()
    assert order2 != CollationElementIterator.NULLORDER
    order3 = it2.next()
    assert order3 != CollationElementIterator.NULLORDER
    assert it2.next() == CollationElementIterator.NULLORDER

    it2.set_offset(3)
    assert it2.get_offset() == 3
    assert it2.previous() == order3
    assert it2.previous() == order2
    assert it2.previous() == order1
    assert it2.previous() == CollationElementIterator.NULLORDER

    assert list(it2) == [order1, order2, order3]

    it2a = coll.create_collation_element_iterator("cha")
    assert isinstance(it2a, CollationElementIterator)

    assert it2a.next() == order1
    assert it2a.next() == order2
    assert it2a.next() == order3
    assert it2a.next() == CollationElementIterator.NULLORDER


def test_create_instance():
    # [1]
    # static Collator *icu::Collator::createInstance(
    #       const Locale &loc,
    #       UErrorCode &err
    # )
    test1 = Collator.create_instance(Locale())
    assert isinstance(test1, RuleBasedCollator)
    rules1 = test1.get_rules()

    # [2]
    # static Collator *icu::Collator::createInstance(UErrorCode &err)
    test2 = Collator.create_instance()
    assert isinstance(test2, RuleBasedCollator)
    rules2 = test2.get_rules()

    assert rules1 == rules2


def test_equals():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # UBool icu::Collator::equals(
    #       const UnicodeString &source,
    #       const UnicodeString &target
    # )
    assert not coll.equals(source, target)
    assert not coll.equals("ABC", target)
    assert not coll.equals(source, "abc")
    assert not coll.equals("ABC", "abc")

    coll.set_attribute(UColAttribute.UCOL_STRENGTH,
                       UColAttributeValue.UCOL_PRIMARY)
    assert coll.equals(source, target)
    assert coll.equals("ABC", target)
    assert coll.equals(source, "abc")
    assert coll.equals("ABC", "abc")


def test_get_attribute():
    coll = Collator.create_instance(Locale.get_canada_french())

    # UColAttributeValue icu::RuleBasedCollator::getAttribute(
    #       UColAttribute attr,
    #       UErrorCode &status
    # )
    assert (coll.get_attribute(UColAttribute.UCOL_FRENCH_COLLATION)
            == UColAttributeValue.UCOL_ON)
    assert (coll.get_attribute(UColAttribute.UCOL_ALTERNATE_HANDLING)
            == UColAttributeValue.UCOL_NON_IGNORABLE)
    assert (coll.get_attribute(UColAttribute.UCOL_CASE_FIRST)
            == UColAttributeValue.UCOL_STRENGTH_LIMIT)
    assert (coll.get_attribute(UColAttribute.UCOL_CASE_LEVEL)
            == UColAttributeValue.UCOL_STRENGTH_LIMIT)
    assert (coll.get_attribute(UColAttribute.UCOL_NORMALIZATION_MODE)
            == UColAttributeValue.UCOL_STRENGTH_LIMIT)
    assert (coll.get_attribute(UColAttribute.UCOL_DECOMPOSITION_MODE)
            == UColAttributeValue.UCOL_STRENGTH_LIMIT)
    assert (coll.get_attribute(UColAttribute.UCOL_STRENGTH)
            == UColAttributeValue.UCOL_TERTIARY)
    assert (coll.get_attribute(UColAttribute.UCOL_NUMERIC_COLLATION)
            == UColAttributeValue.UCOL_STRENGTH_LIMIT)

    # void icu::RuleBasedCollator::setAttribute(
    #       UColAttribute attr,
    #       UColAttributeValue value,
    #       UErrorCode &status
    # )
    coll.set_attribute(UColAttribute.UCOL_STRENGTH,
                       UColAttributeValue.UCOL_PRIMARY)
    assert (coll.get_attribute(UColAttribute.UCOL_STRENGTH)
            == UColAttributeValue.UCOL_PRIMARY)


def test_get_available_locales():
    # [1]
    # static const Locale *icu::Collator::getAvailableLocales(
    #       int32_t &count
    # )
    it = Collator.get_available_locales()
    assert isinstance(it, list)
    assert len(it) > 0
    assert all(isinstance(x, Locale) for x in it)
    assert Locale("ja") in it


def test_get_bound():
    coll = Collator.create_instance(Locale("sh"))
    source = coll.get_sort_key("Smith", -1)

    # static int32_t icu::Collator::getBound(
    #       const uint8_t *source,
    #       int32_t sourceLength,
    #       UColBoundMode boundType,
    #       uint32_t noOfLevels,
    #       uint8_t *result,
    #       int32_t resultLength,
    #       UErrorCode &status
    # )
    result1 = Collator.get_bound(
        source,
        len(source),
        UColBoundMode.UCOL_BOUND_LOWER,
        1)
    result2 = Collator.get_bound(
        source,
        len(source),
        UColBoundMode.UCOL_BOUND_UPPER,
        1)
    assert isinstance(result1, list)
    assert isinstance(result2, list)
    assert len(result1) > 0
    assert len(result2) > 0
    assert all(isinstance(x, int) for x in result1)
    assert all(isinstance(x, int) for x in result2)
    assert result1 < source
    assert result2 > source


def test_get_collation_key():
    coll = Collator.create_instance(Locale.get_english())

    # [1]
    # CollationKey &icu::Collator::getCollationKey(
    #       const char16_t *source,
    #       int32_t sourceLength,
    #       CollationKey &key,
    #       UErrorCode &status
    # )
    key1 = CollationKey()
    result = coll.get_collation_key("ABC", -1, key1)
    assert isinstance(result, CollationKey)
    assert id(result) == id(key1)

    # [2]
    # CollationKey &icu::Collator::getCollationKey(
    #       const UnicodeString &source,
    #       CollationKey &key,
    #       UErrorCode &status
    # )
    key2 = CollationKey()
    result = coll.get_collation_key(UnicodeString("abc"), key2)
    assert isinstance(result, CollationKey)
    assert id(result) == id(key2)

    result = coll.get_collation_key("abc", key2)
    assert isinstance(result, CollationKey)
    assert id(result) == id(key2)

    assert key1.compare_to(key2) == UCollationResult.UCOL_GREATER
    assert key2.compare_to(key1) == UCollationResult.UCOL_LESS


def test_get_display_name():
    default_locale = None
    try:
        default_locale = Locale.get_default()
        display_locale = Locale.get_us()
        Locale.set_default(display_locale)
        object_locale = Locale("ja")

        # [1]
        # static UnicodeString &icu::Collator::getDisplayName(
        #       const Locale &objectLocale,
        #       const Locale &displayLocale,
        #       UnicodeString &name
        # )
        name1 = UnicodeString()
        result = Collator.get_display_name(
            object_locale,
            display_locale,
            name1)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name1)
        assert name1 == "Japanese"

        # [2]
        # static UnicodeString &icu::Collator::getDisplayName(
        #       const Locale &objectLocale,
        #       UnicodeString &name
        # )
        name2 = UnicodeString()
        result = Collator.get_display_name(object_locale, name2)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name2)
        assert name2 == "Japanese"
    finally:
        if default_locale:
            Locale.set_default(default_locale)


def test_get_equivalent_reorder_codes():
    # static int32_t icu::Collator::getEquivalentReorderCodes(
    #       int32_t reorderCode,
    #       int32_t *dest,
    #       int32_t destCapacity,
    #       UErrorCode &status
    # )
    dest = Collator.get_equivalent_reorder_codes(UScriptCode.USCRIPT_KATAKANA)
    assert isinstance(dest, list)
    assert dest == [
        UScriptCode.USCRIPT_HIRAGANA,
        UScriptCode.USCRIPT_KATAKANA,
        UScriptCode.USCRIPT_KATAKANA_OR_HIRAGANA,
    ]


def test_get_functional_equivalent():
    # static Locale icu::Collator::getFunctionalEquivalent(
    #       const char *keyword,
    #       const Locale &locale,
    #       UBool &isAvailable,
    #       UErrorCode &status
    # )
    result, is_available = Collator.get_functional_equivalent(
        "collation",
        Locale.get_japanese())
    assert isinstance(result, Locale)
    assert result == Locale.get_japanese()
    assert is_available


def test_get_keywords():
    # static StringEnumeration *icu::Collator::getKeywords(UErrorCode &status)
    keywords = Collator.get_keywords()
    assert isinstance(keywords, StringEnumeration)
    assert len(keywords) > 0
    assert "collation" in keywords

    # static StringEnumeration *icu::Collator::getKeywordValues(
    #       const char *keyword,
    #       UErrorCode &status
    # )
    values = Collator.get_keyword_values("collation")
    assert isinstance(values, StringEnumeration)
    assert len(values) > 0
    assert "standard" in values

    # static StringEnumeration *icu::Collator::getKeywordValuesForLocale(
    #       const char *keyword,
    #       const Locale &locale,
    #       UBool commonlyUsed,
    #       UErrorCode &status
    # )
    it = Collator.get_keyword_values_for_locale(
        "collation",
        Locale.get_japanese(),
        True)
    assert isinstance(it, StringEnumeration)
    assert list(it) == ['standard', 'unihan', 'emoji', 'eor', 'search']


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 53, reason="ICU4C<53")
def test_get_max_variable():
    coll = Collator.create_instance(Locale.get_japanese())

    # UColReorderCode icu::Collator::getMaxVariable()
    assert (coll.get_max_variable()
            == UColReorderCode.UCOL_REORDER_CODE_PUNCTUATION)

    # uint32_t icu::Collator::getVariableTop(UErrorCode &status)
    weight1 = coll.get_variable_top()
    assert isinstance(weight1, int)

    # Collator& icu::Collator::setMaxVariable(
    #       UColReorderCode group,
    #       UErrorCode &errorCode
    # )
    coll.set_max_variable(UColReorderCode.UCOL_REORDER_CODE_CURRENCY)
    assert (coll.get_max_variable()
            == UColReorderCode.UCOL_REORDER_CODE_CURRENCY)
    weight2 = coll.get_variable_top()
    assert isinstance(weight2, int)

    assert weight1 != weight2


def test_get_reorder_codes():
    coll = Collator.create_instance(Locale.get_japanese())

    # int32_t icu::Collator::getReorderCodes(
    #       int32_t *dest,
    #       int32_t destCapacity,
    #       UErrorCode &status
    # )
    dest = coll.get_reorder_codes()
    assert isinstance(dest, list)
    assert dest == [
        UScriptCode.USCRIPT_LATIN,
        UScriptCode.USCRIPT_KATAKANA,
        UScriptCode.USCRIPT_HAN,
    ]

    # void icu::Collator::setReorderCodes(
    #       const int32_t *reorderCodes,
    #       int32_t reorderCodesLength,
    #       UErrorCode &status
    # )
    reorder_codes = [
        UScriptCode.USCRIPT_HAN,
        UScriptCode.USCRIPT_KATAKANA,
        UScriptCode.USCRIPT_LATIN,
    ]
    coll.set_reorder_codes(reorder_codes, len(reorder_codes))
    dest = coll.get_reorder_codes()
    assert dest == reorder_codes


def test_get_rules():
    coll = Collator.create_instance(Locale.get_japanese())

    # [1]
    # UnicodeString &icu::RuleBasedCollator::getRules()
    rules1 = coll.get_rules()
    assert isinstance(rules1, UnicodeString)

    # [2]
    # void icu::RuleBasedCollator::getRules(
    #       UColRuleOption delta,
    #       UnicodeString &buffer
    # )
    rules2 = UnicodeString()
    coll.get_rules(UColRuleOption.UCOL_TAILORING_ONLY, rules2)
    assert rules2 == rules1

    rules3 = UnicodeString()
    coll.get_rules(UColRuleOption.UCOL_FULL_RULES, rules3)
    assert rules3 != rules1


def test_get_sort_key():
    coll = Collator.create_instance(Locale.get_japanese())

    # [1]
    # int32_t icu::Collator::getSortKey(
    #       const char16_t *source,
    #       int32_t sourceLength,
    #       uint8_t *result,
    #       int32_t resultLength
    # )
    result1 = coll.get_sort_key("ABC", -1)
    assert isinstance(result1, list)
    assert len(result1) > 0
    assert all(isinstance(x, int) for x in result1)

    # [2]
    # int32_t icu::Collator::getSortKey(
    #       const UnicodeString &source,
    #       uint8_t *result,
    #       int32_t resultLength
    # )
    result2 = coll.get_sort_key(UnicodeString("abc"))
    assert isinstance(result2, list)
    assert len(result2) > 0
    assert all(isinstance(x, int) for x in result2)

    result2a = coll.get_sort_key("abc")
    assert result2a == result2

    assert result1 > result2


def test_get_tailored_set():
    coll1 = Collator.create_instance(Locale.get_english())

    # UnicodeSet *icu::Collator::getTailoredSet(UErrorCode &status)
    uniset1 = coll1.get_tailored_set()
    assert isinstance(uniset1, UnicodeSet)
    assert len(uniset1) == 0

    coll2 = Collator.create_instance(Locale.get_japanese())
    uniset2 = coll2.get_tailored_set()
    assert isinstance(uniset2, UnicodeSet)
    assert len(uniset2) > 0


def test_get_version():
    coll = Collator.create_instance(Locale.get_japanese())

    # void icu::Collator::getVersion(UVersionInfo info)
    info = coll.get_version()
    assert isinstance(info, list)
    assert len(info) == 4
    assert all(isinstance(x, int) for x in info)
    assert info[0] > 0


def test_greater():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # UBool icu::Collator::greater(
    #       const UnicodeString &source,
    #       const UnicodeString &target
    # )
    assert coll.greater(source, target)
    assert coll.greater("ABC", target)
    assert coll.greater(source, "abc")
    assert coll.greater("ABC", "abc")

    coll.set_attribute(UColAttribute.UCOL_STRENGTH,
                       UColAttributeValue.UCOL_PRIMARY)
    assert not coll.greater(source, target)
    assert not coll.greater("ABC", target)
    assert not coll.greater(source, "abc")
    assert not coll.greater("ABC", "abc")


def test_greater_or_equal():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # UBool icu::Collator::greaterOrEqual(
    #       const UnicodeString &source,
    #       const UnicodeString &target
    # )
    assert coll.greater_or_equal(source, target)
    assert coll.greater_or_equal("ABC", target)
    assert coll.greater_or_equal(source, "abc")
    assert coll.greater_or_equal("ABC", "abc")

    coll.set_attribute(UColAttribute.UCOL_STRENGTH,
                       UColAttributeValue.UCOL_PRIMARY)
    assert coll.greater_or_equal(source, target)
    assert coll.greater_or_equal("ABC", target)
    assert coll.greater_or_equal(source, "abc")
    assert coll.greater_or_equal("ABC", "abc")


def test_hash_code():
    coll1 = Collator.create_instance(Locale.get_english())
    coll2 = Collator.create_instance(Locale("da", "DK", ""))
    coll3 = Collator.create_instance(Locale.get_english())

    # int32_t icu::Collator::hashCode(void)
    assert coll1.hash_code() != coll2.hash_code()
    assert coll1.hash_code() == coll3.hash_code()
    assert coll2.hash_code() != coll3.hash_code()


def test_operator():
    coll1 = Collator.create_instance(Locale.get_english())
    coll2 = Collator.create_instance(Locale("da", "DK", ""))
    coll3 = Collator.create_instance(Locale.get_english())

    # UBool icu::Collator::operator!=(const Collator &other)
    assert coll1 != coll2

    # UBool icu::Collator::operator==(const Collator &other)
    assert coll1 == coll3
    assert coll2 != coll3


def test_rule_based_collator():
    assert issubclass(RuleBasedCollator, Collator)

    base = Collator.create_instance(Locale.get_japanese())
    rules = base.get_rules()
    assert len(rules) > 0

    # [1]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       UErrorCode &status
    # )
    test1 = RuleBasedCollator(rules)
    assert test1.get_rules() == rules

    test1a = RuleBasedCollator(str(rules))
    assert test1a.get_rules() == rules

    # [2]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       ECollationStrength collationStrength,
    #       UErrorCode &status
    # )
    test2 = RuleBasedCollator(
        rules,
        Collator.ECollationStrength.TERTIARY)
    assert test2.get_rules() == rules

    test2a = RuleBasedCollator(
        str(rules),
        Collator.TERTIARY)
    assert test2a.get_rules() == rules

    # [3]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       UColAttributeValue decompositionMode,
    #       UErrorCode &status
    # )
    test3 = RuleBasedCollator(
        rules,
        UColAttributeValue.UCOL_ON)
    assert test3.get_rules() == rules

    test3a = RuleBasedCollator(
        str(rules),
        UColAttributeValue.UCOL_ON)
    assert test3a.get_rules() == rules

    # [4]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       ECollationStrength collationStrength,
    #       UColAttributeValue decompositionMode,
    #       UErrorCode &status
    # )
    test4 = RuleBasedCollator(
        rules,
        Collator.ECollationStrength.TERTIARY,
        UColAttributeValue.UCOL_ON)
    assert test4.get_rules() == rules

    test4a = RuleBasedCollator(
        str(rules),
        Collator.ECollationStrength.TERTIARY,
        UColAttributeValue.UCOL_ON)
    assert test4a.get_rules() == rules

    # [6]
    # icu::RuleBasedCollator::RuleBasedCollator(const RuleBasedCollator &other)
    test6 = RuleBasedCollator(test1)
    assert test6.get_rules() == rules

    # [7]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const uint8_t *bin,
    #       int32_t length,
    #       const RuleBasedCollator *base,
    #       UErrorCode &status
    # )
    # buffer = test1.clone_binary()
    # assert isinstance(buffer, list)
    # assert len(buffer) > 0
    # root = Collator.create_instance(Locale.get_root())
    # test7 = RuleBasedCollator(buffer, len(buffer), root)
    # rules7 = test7.get_rules()
    # assert rules7 == rules
    pass  # NotImplemented


def test_sort():
    # From https://unicode-org.github.io/icu/userguide/collation/customization/#simple-tailoring-examples
    src = [
        "CUKIĆ RADOJICA", "ČUKIĆ SLOBODAN", "CUKIĆ SVETOZAR", "ČUKIĆ ZORAN",
        "CURIĆ MILOŠ", "ĆURIĆ MILOŠ", "CVRKALJ ÐURO",
    ]
    base = Collator.create_instance(Locale("sr_Latn"))
    rules = base.get_rules() + UnicodeString("& C < č <<< Č < ć <<< Ć")
    coll = RuleBasedCollator(rules)

    def _cmp(_a, _b):
        nonlocal coll
        _result = coll.compare(_a, -1, _b, -1)
        return (
            -1 if _result == UCollationResult.UCOL_LESS
            else 0 if _result == UCollationResult.UCOL_EQUAL
            else 1
        )

    result1 = sorted(src)
    result2 = sorted(src, key=cmp_to_key(_cmp))
    result3 = sorted(src, key=lambda s: coll.get_sort_key(s, -1))
    assert result1 != result2
    assert result1 != result3
    assert result2 == result3
    assert result3 == [
        "CUKIĆ RADOJICA", "CUKIĆ SVETOZAR", "CURIĆ MILOŠ", "CVRKALJ ÐURO",
        "ČUKIĆ SLOBODAN", "ČUKIĆ ZORAN", "ĆURIĆ MILOŠ",
    ]



from functools import partial
from pathlib import Path

import pytest
from icupy import (
    ConstVoidPtr, ICUError, Locale, UErrorCode, UnicodeSet, UnicodeString,
    u_strlen,
    # ucnv
    UCNV_LOCALE_OPTION_STRING, UCNV_VERSION_OPTION_STRING,
    UConverterPlatform, UConverterType, UConverterUnicodeSet,
    ucnv_close, ucnv_compare_names, ucnv_count_aliases, ucnv_count_available,
    ucnv_count_standards, ucnv_detect_unicode_signature,
    ucnv_fix_file_separator, ucnv_flush_cache,
    ucnv_get_alias, ucnv_get_aliases, ucnv_get_available_name,
    ucnv_get_canonical_name, ucnv_get_ccsid, ucnv_get_default_name,
    ucnv_get_display_name, ucnv_get_from_u_call_back, ucnv_get_name,
    ucnv_get_platform, ucnv_get_standard, ucnv_get_standard_name,
    ucnv_get_subst_chars, ucnv_get_to_u_call_back, ucnv_get_type,
    ucnv_get_unicode_set,
    ucnv_is_ambiguous, ucnv_is_fixed_width,
    ucnv_open, ucnv_open_all_names, ucnv_open_ccsid, ucnv_open_package,
    ucnv_open_standard_names, ucnv_reset,
    ucnv_reset_from_unicode, ucnv_reset_to_unicode,
    ucnv_set_fallback, ucnv_set_from_u_call_back, ucnv_set_subst_chars,
    ucnv_set_subst_string, ucnv_set_to_u_call_back, ucnv_uses_fallback,
    # ucnv_err
    UCNV_ESCAPE_C, UCNV_ESCAPE_XML_HEX,
    UCNV_FROM_U_CALLBACK_ESCAPE, UCNV_TO_U_CALLBACK_ESCAPE,
    UConverterCallbackReason,
    UConverterFromUCallbackPtr, UConverterFromUnicodeArgs,
    UConverterToUCallbackPtr, UConverterToUnicodeArgs,
    # ucnv_cb
    ucnv_cb_from_u_write_bytes, ucnv_cb_from_u_write_sub,
    ucnv_cb_to_u_write_sub, ucnv_cb_to_u_write_uchars,
    # uenum
    uenum_close, uenum_count, uenum_next,
)

from . import gc


def test_api():
    assert ucnv_compare_names("UTF-8", "utf8") == 0
    assert ucnv_compare_names("UTF-8", "utf80") != 0

    count = ucnv_count_available()
    assert count > 0
    names = [ucnv_get_available_name(n) for n in range(count)]

    name = UnicodeString("ISO_2022", -1)
    name.append(UCNV_LOCALE_OPTION_STRING, -1).append(Locale.JAPANESE, -1)
    name.append(UCNV_VERSION_OPTION_STRING, -1).append("4", -1)
    encoding = str(name)  # "ISO_2022,locale=ja,version=4"
    assert encoding in names

    count = ucnv_count_aliases(encoding)
    assert count > 0
    aliases1 = [ucnv_get_alias(encoding, n) for n in range(count)]
    assert "JIS8" in aliases1
    aliases2 = ucnv_get_aliases(encoding)
    assert isinstance(aliases2, list)
    assert sorted(aliases1) == sorted(aliases2)

    count = ucnv_count_standards()
    assert count > 0
    standards = [ucnv_get_standard(n) for n in range(count)]
    assert "IANA" in standards

    # const char *ucnv_detectUnicodeSignature(const char *source,
    #                                         int32_t sourceLength,
    #                                         int32_t *signatureLength,
    #                                         UErrorCode *pErrorCode
    # )
    source = b"\xEF\xBB\xBF\x41\x42\x43"
    result = ucnv_detect_unicode_signature(source, len(source))
    assert result == "UTF-8"
    result = ucnv_detect_unicode_signature(source, 2)
    assert result is None

    result = ucnv_get_canonical_name("utf8", "IANA")
    assert result == "UTF-8"

    result = ucnv_get_default_name()
    assert isinstance(result, str)
    assert len(result) > 0

    result = ucnv_get_standard_name("ibm-943_P15A-2003", "IANA")
    assert result == "Shift_JIS"

    with gc(ucnv_open("ibm-943"), ucnv_close) as cnv:
        assert ucnv_get_ccsid(cnv) == 943

        result = ucnv_get_display_name(cnv, Locale.ENGLISH)
        assert result == "ibm-943_P130-1999"

        # UnicodeString::UnicodeString(const char *src,
        #                              int32_t srcLength,
        #                              UConverter *cnv,
        #                              UErrorCode &errorCode
        # )
        sjis = b"\x61\x5B\x5C\xB1\x87\x54"
        utf8 = sjis.decode("cp932")
        test1 = UnicodeString(sjis, -1, cnv)  # ibm-943 to utf-8
        source = test1.extract()
        assert isinstance(source, str)

        # void ucnv_fixFileSeparator(const UConverter *cnv,
        #                            UChar *source,
        #                            int32_t sourceLen
        # )
        result = ucnv_fix_file_separator(cnv, source, u_strlen(source))
        assert isinstance(result, str)
        assert source != utf8
        assert result == utf8

        # int32_t UnicodeString::extract(char *dest,
        #                                int32_t destCapacity,
        #                                UConverter *cnv,
        #                                UErrorCode &errorCode
        # )
        dest = test1.extract(cnv)  # utf-8 to ibm-943
        assert isinstance(dest, bytes)
        assert dest.decode("cp932") == utf8

        assert ucnv_get_name(cnv) == "ibm-943_P130-1999"
        assert ucnv_get_platform(cnv) == UConverterPlatform.UCNV_IBM
        assert ucnv_get_subst_chars(cnv) == b"\xFC\xFC"
        assert ucnv_get_type(cnv) == UConverterType.UCNV_MBCS

        uniset = UnicodeSet()
        ucnv_get_unicode_set(cnv,
                             uniset.to_uset(),
                             UConverterUnicodeSet.UCNV_ROUNDTRIP_SET)
        assert uniset.size() > 0

        assert ucnv_is_ambiguous(cnv)
        assert not ucnv_is_fixed_width(cnv)
        assert not ucnv_uses_fallback(cnv)
        ucnv_set_fallback(cnv, True)
        assert ucnv_uses_fallback(cnv)

        test2 = UnicodeString("a\uFF71b\U0001f338c", -1)
        dest = test2.extract(cnv)  # utf-8 to ibm-943
        # dest.decode("cp932") raises UnicodeDecodeError.
        assert dest == b"a\xB1b\xFC\xFCc"

        sub = b"?"
        ucnv_set_subst_chars(cnv, sub, len(sub))
        assert ucnv_get_subst_chars(cnv) == sub

        dest = test2.extract(cnv)  # utf-8 to ibm-943
        assert dest.decode("cp932") == "a\uFF71b?c"

        sub = "<??>"
        ucnv_set_subst_string(cnv, sub, len(sub))
        assert ucnv_get_subst_chars(cnv) == sub.encode()

        dest = test2.extract(cnv)  # utf-8 to ibm-943
        assert dest.decode("cp932") == "a\uFF71b<??>c"

        ucnv_reset(cnv)
        ucnv_reset_from_unicode(cnv)
        ucnv_reset_to_unicode(cnv)

    en = ucnv_open_all_names()
    assert uenum_count(en) > 0
    names = iter(partial(uenum_next, en), None)
    assert encoding in names
    uenum_close(en)

    en = ucnv_open_standard_names("ibm-943_P15A-2003", "IANA")
    assert uenum_count(en) > 0
    standards = iter(partial(uenum_next, en), None)
    assert "Shift_JIS" in standards
    uenum_close(en)

    with gc(ucnv_open_ccsid(943, UConverterPlatform.UCNV_IBM),
            ucnv_close) as cnv:
        assert ucnv_get_ccsid(cnv) == 943

        result = ucnv_get_display_name(cnv, Locale.ENGLISH)
        assert result == "ibm-943_P130-1999"

    ucnv_flush_cache()


def test_open_package():
    path = Path(__file__).joinpath("..", "testdata").resolve()
    cnv = None
    try:
        cnv = ucnv_open_package(str(path), "test3")
    except ICUError as ex:
        if ex.args[0] != UErrorCode.U_FILE_ACCESS_ERROR:
            raise
        pytest.skip("testdata.dat is not found (not an error). "
                    "You need to build a test data from the source. "
                    "See also <icu4c>/icu/source/test/testdata/")

    with gc(cnv, ucnv_close) as cnv:
        assert ucnv_get_name(cnv) == "test3"
        assert ucnv_get_subst_chars(cnv) == b"\xFF"


def test_set_from_u_call_back():
    result1, result2, result3 = [], [], []

    # void (* UConverterFromUCallback)(const void *context,
    #                                  UConverterFromUnicodeArgs *args,
    #                                  const UChar *codeUnits,
    #                                  int32_t length,
    #                                  UChar32 codePoint,
    #                                  UConverterCallbackReason reason,
    #                                  UErrorCode *pErrorCode)
    def _from_u_callback1(_context: object, _args: UConverterFromUnicodeArgs,
                          _code_units: str, _length: int, _code_point: int,
                          _reason: UConverterCallbackReason,
                          _error_code: UErrorCode) -> UErrorCode:
        assert _context is None
        nonlocal result1
        result1.append((_reason, _error_code, _code_units))
        assert ucnv_get_name(_args.converter) == "ibm-943_P15A-2003"
        if _reason == UConverterCallbackReason.UCNV_UNASSIGNED:
            _source = "?"
            ucnv_cb_from_u_write_bytes(_args, _source, len(_source), 0)
            _error_code = UErrorCode.U_ZERO_ERROR
        return _error_code

    def _from_u_callback2(_context: object, _args: UConverterFromUnicodeArgs,
                          _code_units: str, _length: int, _code_point: int,
                          _reason: UConverterCallbackReason,
                          _error_code: UErrorCode) -> UErrorCode:
        assert isinstance(_context, str)
        assert _context == "foo bar baz"
        nonlocal result2
        result2.append((_reason, _error_code, _code_units))
        assert ucnv_get_name(_args.converter) == "ibm-943_P15A-2003"
        if _reason == UConverterCallbackReason.UCNV_UNASSIGNED:
            ucnv_cb_from_u_write_sub(_args, 0)
            _error_code = UErrorCode.U_ZERO_ERROR
        return _error_code

    def _from_u_callback3(_context: object, _args: UConverterFromUnicodeArgs,
                          _code_units: str, _length: int, _code_point: int,
                          _reason: UConverterCallbackReason,
                          _error_code: UErrorCode) -> UErrorCode:
        assert isinstance(_context, dict)
        assert _context["key"] == "foo bar baz"
        nonlocal result3
        result3.append((_reason, _error_code, _code_units))
        assert ucnv_get_name(_args.converter) == "ibm-943_P15A-2003"
        if _reason == UConverterCallbackReason.UCNV_UNASSIGNED:
            if _code_point >= 0xffff:
                _fmt = "\\U{:08x}"
            elif _code_point >= 0xff:
                _fmt = "\\u{:04x}"
            else:
                _fmt = "\\x{:02x}"
            _source = _fmt.format(_code_point)
            ucnv_cb_from_u_write_bytes(_args, _source, len(_source), 0)
            _error_code = UErrorCode.U_ZERO_ERROR
        return _error_code

    with gc(ucnv_open("ibm-943c"), ucnv_close) as cnv:
        assert ucnv_get_subst_chars(cnv) == b"\xfc\xfc"

        test1 = UnicodeString("a\uFF71b\U0001f338c", -1)

        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert dest == b"a\xb1b\xfc\xfcc"

        # void ucnv_getFromUCallBack(const UConverter *converter,
        #                            UConverterFromUCallback *action,
        #                            const void **context
        # )
        old_action1, old_context1 = ucnv_get_from_u_call_back(cnv)

        # void ucnv_setFromUCallBack(UConverter *converter,
        #                            UConverterFromUCallback newAction,
        #                            const void *newContext,
        #                            UConverterFromUCallback *oldAction,
        #                            const void **oldContext,
        #                            UErrorCode *err
        # )
        context2 = ConstVoidPtr(None)
        ucnv_set_from_u_call_back(
            cnv,
            UCNV_FROM_U_CALLBACK_ESCAPE,
            context2)  # %UXXXX
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert dest == b"a\xb1b%UD83C%UDF38c"

        context3 = ConstVoidPtr(UCNV_ESCAPE_C)
        old_action3, old_context3 = ucnv_set_from_u_call_back(
            cnv,
            UCNV_FROM_U_CALLBACK_ESCAPE,
            context3)  # \\uXXXX
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert dest == b"a\xb1b\\U0001F338c"

        context4 = ConstVoidPtr(UCNV_ESCAPE_XML_HEX)
        ucnv_set_from_u_call_back(
            cnv,
            UCNV_FROM_U_CALLBACK_ESCAPE,
            context4)  # &#xXXXX;
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert dest == b"a\xb1b&#x1F338;c"

        action5 = UConverterFromUCallbackPtr(_from_u_callback1)
        context5 = ConstVoidPtr(None)
        ucnv_set_from_u_call_back(
            cnv,
            action5,
            context5)
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert len(result1) > 0
        assert dest == b"a\xb1b?c"

        action6 = UConverterFromUCallbackPtr(_from_u_callback2)
        context6 = ConstVoidPtr("foo bar baz")
        ucnv_set_from_u_call_back(
            cnv,
            action6,
            context6)
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert len(result2) > 0
        assert dest == b"a\xb1b\xfc\xfcc"

        action7 = UConverterFromUCallbackPtr(_from_u_callback3)
        d = {"key": "foo bar baz"}
        context7 = ConstVoidPtr(d)
        old_action7, old_context7 = ucnv_set_from_u_call_back(
            cnv,
            action7,
            context7)
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert len(result3) > 0
        assert dest == b"a\xb1b\\U0001f338c"

        ucnv_set_from_u_call_back(
            cnv,
            old_action7,
            old_context7)
        result1.clear()
        result2.clear()
        result3.clear()
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert len(result2) > 0
        assert dest == b"a\xb1b\xfc\xfcc"

        ucnv_set_from_u_call_back(
            cnv,
            old_action3,
            old_context3)
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert dest == b"a\xb1b%UD83C%UDF38c"

        ucnv_set_from_u_call_back(
            cnv,
            old_action1,
            old_context1)
        dest = test1.extract(cnv)  # utf-8 to ibm-943c
        assert dest == b"a\xb1b\xfc\xfcc"


def test_set_to_u_call_back():
    result1, result2, result3 = [], [], []

    # void (* UConverterToUCallback)(const void *context,
    #                                UConverterToUnicodeArgs *args,
    #                                const char *codeUnits,
    #                                int32_t length,
    #                                UConverterCallbackReason reason,
    #                                UErrorCode *pErrorCode)
    def _to_u_callback1(_context: object, _args: UConverterToUnicodeArgs,
                        _code_units: bytes, _length: int,
                        _reason: UConverterCallbackReason,
                        _error_code: UErrorCode) -> UErrorCode:
        assert _context is None
        nonlocal result1
        result1.append((_reason, _error_code, _code_units))
        assert ucnv_get_name(_args.converter) == "UTF-8"
        if _reason == UConverterCallbackReason.UCNV_ILLEGAL:
            ucnv_cb_to_u_write_sub(_args, 0)
            _error_code = UErrorCode.U_ZERO_ERROR
        return _error_code

    def _to_u_callback2(_context: object, _args: UConverterToUnicodeArgs,
                        _code_units: bytes, _length: int,
                        _reason: UConverterCallbackReason,
                        _error_code: UErrorCode) -> UErrorCode:
        assert isinstance(_context, str)
        assert _context == "foo bar baz"
        nonlocal result2
        result2.append((_reason, _error_code, _code_units))
        assert ucnv_get_name(_args.converter) == "UTF-8"
        if _reason == UConverterCallbackReason.UCNV_ILLEGAL:
            _source = "".join(["\\x{:02x}".format(x) for x in _code_units])
            ucnv_cb_to_u_write_uchars(_args, _source, len(_source), 0)
            _error_code = UErrorCode.U_ZERO_ERROR
        return _error_code

    def _to_u_callback3(_context: object, _args: UConverterToUnicodeArgs,
                        _code_units: bytes, _length: int,
                        _reason: UConverterCallbackReason,
                        _error_code: UErrorCode) -> UErrorCode:
        assert isinstance(_context, dict)
        assert _context["key"] == "foo bar baz"
        nonlocal result3
        result3.append((_reason, _error_code, _code_units))
        assert ucnv_get_name(_args.converter) == "UTF-8"
        if _reason == UConverterCallbackReason.UCNV_ILLEGAL:
            _source = "".join(["%x{:02X}".format(x) for x in _code_units])
            ucnv_cb_to_u_write_uchars(_args, _source, len(_source), 0)
            _error_code = UErrorCode.U_ZERO_ERROR
        return _error_code

    with gc(ucnv_open("utf-8"), ucnv_close) as cnv:
        assert ucnv_get_subst_chars(cnv) == b"\xef\xbf\xbd"

        # Illegal UTF-8 characters
        utf8 = (b"\x61"
                b"\xfe"  # Impossible bytes
                b"\xc0\xaf"  # Overlong sequences
                b"\xed\xa0\x80"  # Illegal code positions
                b"\x62")

        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        dest = test1.extract()
        assert (dest.encode() == b"a"
                                 b"\xef\xbf\xbd"
                                 b"\xef\xbf\xbd\xef\xbf\xbd"
                                 b"\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd"
                                 b"b")

        # void ucnv_getToUCallBack(const UConverter *converter,
        #                          UConverterToUCallback *action,
        #                          const void **context
        # )
        old_action1, old_context1 = ucnv_get_to_u_call_back(cnv)

        # void ucnv_setToUCallBack(UConverter *converter,
        #                          UConverterToUCallback newAction,
        #                          const void *newContext,
        #                          UConverterToUCallback *oldAction,
        #                          const void **oldContext,
        #                          UErrorCode *err
        # )
        context2 = ConstVoidPtr(None)
        ucnv_set_to_u_call_back(
            cnv,
            UCNV_TO_U_CALLBACK_ESCAPE,
            context2)  # %XNN
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        dest = test1.extract()
        assert (dest == "a"
                        "%XFE"
                        "%XC0%XAF"
                        "%XED%XA0%X80"
                        "b")

        context3 = ConstVoidPtr(UCNV_ESCAPE_C)
        old_action3, old_context3 = ucnv_set_to_u_call_back(
            cnv,
            UCNV_TO_U_CALLBACK_ESCAPE,
            context3)  # \\xXX
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        dest = test1.extract()
        assert (dest == "a"
                        "\\xFE"
                        "\\xC0\\xAF"
                        "\\xED\\xA0\\x80"
                        "b")

        context4 = ConstVoidPtr(UCNV_ESCAPE_XML_HEX)
        ucnv_set_to_u_call_back(
            cnv,
            UCNV_TO_U_CALLBACK_ESCAPE,
            context4)  # &#xXX;
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        dest = test1.extract()
        assert (dest == "a"
                        "&#xFE;"
                        "&#xC0;&#xAF;"
                        "&#xED;&#xA0;&#x80;"
                        "b")

        action5 = UConverterToUCallbackPtr(_to_u_callback1)
        context5 = ConstVoidPtr(None)
        ucnv_set_to_u_call_back(
            cnv,
            action5,
            context5)
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        assert len(result1) > 0
        dest = test1.extract()
        assert (dest.encode() == b"a"
                                 b"\xef\xbf\xbd"
                                 b"\xef\xbf\xbd\xef\xbf\xbd"
                                 b"\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd"
                                 b"b")

        action6 = UConverterToUCallbackPtr(_to_u_callback2)
        context6 = ConstVoidPtr("foo bar baz")
        ucnv_set_to_u_call_back(
            cnv,
            action6,
            context6)
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        assert len(result2) > 0
        dest = test1.extract()
        assert (dest == "a"
                        "\\xfe"
                        "\\xc0\\xaf"
                        "\\xed\\xa0\\x80"
                        "b")

        action7 = UConverterToUCallbackPtr(_to_u_callback3)
        d = {"key": "foo bar baz"}
        context7 = ConstVoidPtr(d)
        old_action7, old_context7 = ucnv_set_to_u_call_back(
            cnv,
            action7,
            context7)
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        assert len(result3) > 0
        dest = test1.extract()
        assert (dest == "a"
                        "%xFE"
                        "%xC0%xAF"
                        "%xED%xA0%x80"
                        "b")

        ucnv_set_to_u_call_back(
            cnv,
            old_action7,
            old_context7)
        result1.clear()
        result2.clear()
        result3.clear()
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        assert len(result2) > 0
        dest = test1.extract()
        assert (dest == "a"
                        "\\xfe"
                        "\\xc0\\xaf"
                        "\\xed\\xa0\\x80"
                        "b")

        ucnv_set_to_u_call_back(
            cnv,
            old_action3,
            old_context3)
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        dest = test1.extract()
        assert (dest == "a"
                        "%XFE"
                        "%XC0%XAF"
                        "%XED%XA0%X80"
                        "b")

        ucnv_set_to_u_call_back(
            cnv,
            old_action1,
            old_context1)
        test1 = UnicodeString(utf8, -1, cnv)  # utf-8 to utf-8
        dest = test1.extract()
        assert (dest.encode() == b"a"
                                 b"\xef\xbf\xbd"
                                 b"\xef\xbf\xbd\xef\xbf\xbd"
                                 b"\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd"
                                 b"b")



import copy

import pytest
from icupy import (
    DateFormat, DateFormatSymbols, FieldPosition, FieldPositionIterator,
    Format, Formattable, GregorianCalendar, ICUError, Locale,
    ParsePosition, SimpleDateFormat, TimeZone, TimeZoneFormat,
    UCalendarMonths, UDateFormatField, UErrorCode, U_ICU_VERSION_MAJOR_NUM,
    UnicodeString,
)


def test_api():
    fmt1 = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        Locale.get_english())
    fmt2 = fmt1.clone()
    fmt3 = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        Locale.get_french())

    # UBool icu::Format::operator!=(const Format &other)
    assert not (fmt1 != fmt2)
    assert fmt1 != fmt3
    assert fmt2 != fmt3

    # UBool icu::SimpleDateFormat::operator==(const Format &other)
    assert fmt1 == fmt2
    assert not (fmt1 == fmt3)
    assert not (fmt2 == fmt3)

    # UnicodeString &icu::SimpleDateFormat::toLocalizedPattern(
    #       UnicodeString &result,
    #       UErrorCode &status
    # )
    result = UnicodeString()
    string = fmt2.to_localized_pattern(result)
    assert isinstance(string, UnicodeString)
    assert id(result) == id(string)
    assert result == "yyyy.MM.dd G 'at' HH:mm:ss z"

    # void icu::SimpleDateFormat::applyLocalizedPattern(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    fmt2.apply_localized_pattern(UnicodeString("MMMMdHmm"))
    assert fmt2.to_localized_pattern(result) == "MMMMdHmm"

    fmt2.apply_localized_pattern("yyyy.MM.dd G 'at' HH:mm:ss z")
    assert fmt2.to_localized_pattern(result) == "yyyy.MM.dd G 'at' HH:mm:ss z"

    # UnicodeString &icu::SimpleDateFormat::toPattern(UnicodeString &result)
    string = fmt2.to_pattern(result)
    assert isinstance(string, UnicodeString)
    assert id(result) == id(string)
    assert result == "yyyy.MM.dd G 'at' HH:mm:ss z"

    # void icu::SimpleDateFormat::applyPattern(const UnicodeString &pattern)
    fmt2.apply_pattern(UnicodeString("MMMMdHmm"))
    assert fmt2.to_pattern(result) == "MMMMdHmm"

    fmt2.apply_pattern("yyyy.MM.dd G 'at' HH:mm:ss z")
    assert fmt2.to_pattern(result) == "yyyy.MM.dd G 'at' HH:mm:ss z"

    # UDate icu::SimpleDateFormat::get2DigitYearStart(UErrorCode &status)
    d = fmt2.get_2digit_year_start()
    assert isinstance(d, float)

    # void icu::SimpleDateFormat::set2DigitYearStart(
    #       UDate d,
    #       UErrorCode &status
    # )
    fmt2.set_2digit_year_start(d)

    # const DateFormatSymbols *
    # icu::SimpleDateFormat::getDateFormatSymbols(void)
    sym = fmt2.get_date_format_symbols()
    assert isinstance(sym, DateFormatSymbols)

    # void icu::SimpleDateFormat::setDateFormatSymbols(
    #       const DateFormatSymbols &newFormatSymbols
    # )
    new_format_symbols = DateFormatSymbols(Locale.get_french())
    assert new_format_symbols != sym
    fmt2.set_date_format_symbols(new_format_symbols)
    assert fmt2.get_date_format_symbols() == new_format_symbols


def test_clone():
    fmt1 = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        Locale.get_english())

    # SimpleDateFormat *icu::SimpleDateFormat::clone()
    fmt2 = fmt1.clone()
    assert isinstance(fmt2, SimpleDateFormat)
    assert fmt2 == fmt1

    fmt3 = copy.copy(fmt1)
    assert fmt3 == fmt1

    fmt4 = copy.deepcopy(fmt1)
    assert fmt4 == fmt1


def test_format():
    locale = Locale.get_english()
    tz = TimeZone.create_time_zone("PST")
    cal = GregorianCalendar(tz, locale)
    cal.set(1996, UCalendarMonths.UCAL_JULY, 10, 15, 8, 56)
    date = cal.get_time()
    obj = Formattable(date, Formattable.IS_DATE)
    fmt = SimpleDateFormat("yyyy.MM.dd G 'at' HH:mm:ss z", locale)
    fmt.set_time_zone(tz)

    # [2]
    # UnicodeString &icu::SimpleDateFormat::format(
    #       Calendar &cal,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos
    # )
    append_to = UnicodeString()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt.format(cal, append_to, pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [4]
    # UnicodeString &icu::SimpleDateFormat::format(
    #       Calendar &cal,
    #       UnicodeString &appendTo,
    #       FieldPositionIterator *posIter,
    #       UErrorCode &status
    # )
    append_to.remove()
    pos_iter = FieldPositionIterator()
    result = fmt.format(cal, append_to, pos_iter)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    fp = FieldPosition()
    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_YEAR_FIELD
    assert fp.get_begin_index() == 0
    assert fp.get_end_index() == 4

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_MONTH_FIELD

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_DATE_FIELD

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_ERA_FIELD

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_HOUR_OF_DAY0_FIELD

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_MINUTE_FIELD

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_SECOND_FIELD

    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_TIMEZONE_FIELD

    assert not pos_iter.next(fp)

    append_to.remove()
    result = fmt.format(cal, append_to, None)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [5]
    # UnicodeString &icu::DateFormat::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos,
    #       UErrorCode &status
    # )
    append_to.remove()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt.format(obj, append_to, pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [8]
    # UnicodeString &icu::DateFormat::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPositionIterator *posIter,
    #       UErrorCode &status
    # )
    append_to.remove()
    pos_iter = FieldPositionIterator()
    result = fmt.format(obj, append_to, pos_iter)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    fp = FieldPosition()
    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_YEAR_FIELD
    assert fp.get_begin_index() == 0
    assert fp.get_end_index() == 4

    append_to.remove()
    result = fmt.format(obj, append_to, None)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [9]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to.remove()
    result = fmt.format(obj, append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [10]
    # UnicodeString &icu::DateFormat::format(
    #       UDate date,
    #       UnicodeString &appendTo
    # )
    append_to.remove()
    result = fmt.format(date, append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [11]
    # UnicodeString &icu::DateFormat::format(
    #       UDate date,
    #       UnicodeString &appendTo,
    #       FieldPosition &fieldPosition
    # )
    append_to.remove()
    field_position = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt.format(date, append_to, field_position)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    # [12]
    # UnicodeString &icu::DateFormat::format(
    #       UDate date,
    #       UnicodeString &appendTo,
    #       FieldPositionIterator *posIter,
    #       UErrorCode &status
    # )
    append_to.remove()
    pos_iter = FieldPositionIterator()
    result = fmt.format(date, append_to, pos_iter)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"

    fp = FieldPosition()
    assert pos_iter.next(fp)
    assert fp.get_field() == UDateFormatField.UDAT_YEAR_FIELD
    assert fp.get_begin_index() == 0
    assert fp.get_end_index() == 4

    append_to.remove()
    result = fmt.format(date, append_to, None)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1996.07.10 AD at 15:08:56 PDT"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 53, reason="ICU4C<53")
def test_get_context():
    from icupy import UDisplayContext, UDisplayContextType

    fmt = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        Locale.get_english())

    # UDisplayContext icu::DateFormat::getContext(
    #       UDisplayContextType type,
    #       UErrorCode &status
    # )
    assert (fmt.get_context(UDisplayContextType.UDISPCTX_TYPE_CAPITALIZATION)
            == UDisplayContext.UDISPCTX_CAPITALIZATION_NONE)

    # void icu::SimpleDateFormat::setContext(
    #       UDisplayContext value,
    #       UErrorCode &status
    # )
    fmt.set_context(
        UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE)
    assert (fmt.get_context(UDisplayContextType.UDISPCTX_TYPE_CAPITALIZATION)
            == UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 50, reason="ICU4C<50")
def test_get_time_zone_format():
    fmt = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        Locale.get_english())

    # *ICU 49 technology preview*
    # const TimeZoneFormat *icu::SimpleDateFormat::getTimeZoneFormat(void)
    tzf = fmt.get_time_zone_format()
    assert isinstance(tzf, TimeZoneFormat)

    tzf2 = TimeZoneFormat.create_instance(Locale.get_french())
    assert tzf2 != tzf

    # *ICU 49 technology preview*
    # void icu::SimpleDateFormat::setTimeZoneFormat(
    #       const TimeZoneFormat &newTimeZoneFormat
    # )
    fmt.set_time_zone_format(tzf2)
    assert fmt.get_time_zone_format() == tzf2


def test_parse():
    locale = Locale.get_english()
    tz = TimeZone.create_time_zone("PST")
    fmt = SimpleDateFormat("yyyy.MM.dd G 'at' HH:mm:ss z", locale)
    date = 837036536000.0  # 1996-07-10T15:08:56-07:00

    # [1]
    # void icu::SimpleDateFormat::parse(
    #       const UnicodeString &text,
    #       Calendar &cal,
    #       ParsePosition &pos
    # )
    cal = GregorianCalendar(tz, locale)
    cal.clear()
    pos = ParsePosition(0)
    fmt.parse(UnicodeString("1996.07.10 AD at 15:08:56 PDT"), cal, pos)
    assert pos.get_error_index() == -1
    assert cal.get_time() == date

    cal.clear()
    pos = ParsePosition(0)
    fmt.parse("1996.07.10 AD at 15:08:56 PDT", cal, pos)
    assert pos.get_error_index() == -1
    assert cal.get_time() == date

    # [3]
    # UDate icu::DateFormat::parse(
    #       const UnicodeString &text,
    #       ParsePosition &pos
    # )
    pos = ParsePosition(0)
    result = fmt.parse(UnicodeString("1996.07.10 AD at 15:08:56 PDT"), pos)
    assert pos.get_error_index() == -1
    assert isinstance(result, float)
    assert result == date

    pos = ParsePosition(0)
    result = fmt.parse("1996.07.10 AD at 15:08:56 PDT", pos)
    assert pos.get_error_index() == -1
    assert isinstance(result, float)
    assert result == date

    # [4]
    # UDate icu::DateFormat::parse(
    #       const UnicodeString &text,
    #       UErrorCode &status
    # )
    result = fmt.parse(UnicodeString("1996.07.10 AD at 15:08:56 PDT"))
    assert isinstance(result, float)
    assert result == date

    result = fmt.parse("1996.07.10 AD at 15:08:56 PDT")
    assert isinstance(result, float)
    assert result == date

    with pytest.raises(ICUError) as exc_info:
        _ = fmt.parse("1996.07.10 AD at 15:08:56, PDT")
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR


def test_parse_object():
    fmt = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        Locale.get_english())
    date = 837036536000.0  # 1996-07-10T15:08:56-07:00

    # void icu::DateFormat::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       ParsePosition &parse_pos
    # )
    result = Formattable()
    parse_pos = ParsePosition(0)
    fmt.parse_object(
        UnicodeString("1996.07.10 AD at 15:08:56 PDT"),
        result,
        parse_pos)
    assert parse_pos.get_error_index() == -1
    assert result.get_type() == Formattable.DATE
    assert result.get_date() == date

    result = Formattable()
    parse_pos = ParsePosition(0)
    fmt.parse_object(
        "1996.07.10 AD at 15:08:56 PDT",
        result,
        parse_pos)
    assert parse_pos.get_error_index() == -1
    assert result.get_type() == Formattable.DATE
    assert result.get_date() == date

    # [2]
    # void icu::Format::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       UErrorCode &status
    # )
    result = Formattable()
    fmt.parse_object(
        UnicodeString("1996.07.10 AD at 15:08:56 PDT"),
        result)
    assert result.get_type() == Formattable.DATE
    assert result.get_date() == date

    result = Formattable()
    fmt.parse_object(
        "1996.07.10 AD at 15:08:56 PDT",
        result)
    assert result.get_type() == Formattable.DATE
    assert result.get_date() == date


def test_simple_date_format():
    assert issubclass(DateFormat, Format)
    assert issubclass(SimpleDateFormat, DateFormat)
    locale = Locale.get_english()

    # [1]
    # icu::SimpleDateFormat::SimpleDateFormat(UErrorCode &status)
    fmt1 = SimpleDateFormat()

    # [2]
    # icu::SimpleDateFormat::SimpleDateFormat(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    fmt2 = SimpleDateFormat(UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"))
    fmt2a = SimpleDateFormat("yyyy.MM.dd G 'at' HH:mm:ss z")
    assert fmt2 == fmt2a
    assert fmt2 != fmt1

    # [3]
    # icu::SimpleDateFormat::SimpleDateFormat(
    #       const UnicodeString &pattern,
    #       const UnicodeString &override,
    #       UErrorCode &status
    # )
    fmt3 = SimpleDateFormat(
        UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"),
        UnicodeString("y=hebr;d=thai;s=arab"))
    fmt3a = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        UnicodeString("y=hebr;d=thai;s=arab"))
    fmt3b = SimpleDateFormat(
        UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"),
        "y=hebr;d=thai;s=arab")
    fmt3c = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        "y=hebr;d=thai;s=arab")
    assert fmt3 == fmt3a == fmt3b == fmt3c
    assert fmt3 != fmt1

    # [4]
    # icu::SimpleDateFormat::SimpleDateFormat(
    #       const UnicodeString &pattern,
    #       const Locale &locale,
    #       UErrorCode &status
    # )
    fmt4 = SimpleDateFormat(
        UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"),
        locale)
    fmt4a = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        locale)
    assert fmt4 == fmt4a
    assert fmt4 != fmt1

    # [5]
    # icu::SimpleDateFormat::SimpleDateFormat(
    #       const UnicodeString &pattern,
    #       const UnicodeString &override,
    #       const Locale &locale,
    #       UErrorCode &status
    # )
    fmt5 = SimpleDateFormat(
        UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"),
        UnicodeString("y=hebr;d=thai;s=arab"),
        locale)
    fmt5a = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        UnicodeString("y=hebr;d=thai;s=arab"),
        locale)
    fmt5b = SimpleDateFormat(
        UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"),
        "y=hebr;d=thai;s=arab",
        locale)
    fmt5c = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        "y=hebr;d=thai;s=arab",
        locale)
    assert fmt5 == fmt5a == fmt5b == fmt5c
    assert fmt5 != fmt1

    # [7]
    # icu::SimpleDateFormat::SimpleDateFormat(
    #       const UnicodeString &pattern,
    #       const DateFormatSymbols &formatData,
    #       UErrorCode &status
    # )
    format_data = DateFormatSymbols(locale)
    fmt7 = SimpleDateFormat(
        UnicodeString("yyyy.MM.dd G 'at' HH:mm:ss z"),
        format_data)
    fmt7a = SimpleDateFormat(
        "yyyy.MM.dd G 'at' HH:mm:ss z",
        format_data)
    assert fmt7 == fmt7a
    assert fmt7 != fmt1

    # [8]
    # icu::SimpleDateFormat::SimpleDateFormat(const SimpleDateFormat &)
    fmt8 = SimpleDateFormat(fmt2)
    assert fmt8 == fmt2
    assert fmt8 != fmt1



import copy

import pytest
from icupy import (
    DateTimePatternGenerator, Locale, StringEnumeration,
    UDateTimePatternConflict, UDateTimePatternField,
    UDateTimePatternMatchOptions, U_ICU_VERSION_MAJOR_NUM, UnicodeString,
)


def test_api():
    gen1 = DateTimePatternGenerator.create_instance(Locale("en_US"))
    gen2 = DateTimePatternGenerator.create_instance("en_US")
    gen3 = DateTimePatternGenerator.create_instance(
        Locale("en_US@calendar=japanese"))

    # UBool icu::DateTimePatternGenerator::operator!=(
    #       const DateTimePatternGenerator &other
    # )
    assert not (gen1 != gen2)
    assert gen1 != gen3
    assert gen2 != gen3

    # UBool icu::DateTimePatternGenerator::operator==(
    #       const DateTimePatternGenerator &other
    # )
    assert gen1 == gen2
    assert not (gen1 == gen3)
    assert not (gen2 == gen3)

    # const UnicodeString &icu::DateTimePatternGenerator::getAppendItemFormat(
    #       UDateTimePatternField field
    # )
    result = gen2.get_append_item_format(
        UDateTimePatternField.UDATPG_ERA_FIELD)
    assert isinstance(result, UnicodeString)
    assert result == "{0} {1}"

    # void icu::DateTimePatternGenerator::setAppendItemFormat(
    #       UDateTimePatternField field,
    #       const UnicodeString &value
    # )
    gen2.set_append_item_format(
        UDateTimePatternField.UDATPG_ERA_FIELD,
        UnicodeString("{0}, {1}"))
    assert gen2.get_append_item_format(
        UDateTimePatternField.UDATPG_ERA_FIELD) == "{0}, {1}"

    gen2.set_append_item_format(
        UDateTimePatternField.UDATPG_ERA_FIELD,
        "{0} {1}")
    assert gen2.get_append_item_format(
        UDateTimePatternField.UDATPG_ERA_FIELD) == "{0} {1}"

    # const UnicodeString &icu::DateTimePatternGenerator::getAppendItemName(
    #       UDateTimePatternField field
    # )
    result = gen2.get_append_item_name(
        UDateTimePatternField.UDATPG_ERA_FIELD)
    assert isinstance(result, UnicodeString)
    assert result == "era"

    # void icu::DateTimePatternGenerator::setAppendItemName(
    #       UDateTimePatternField field,
    #       const UnicodeString &value
    # )
    gen2.set_append_item_name(
        UDateTimePatternField.UDATPG_ERA_FIELD,
        UnicodeString("epoch"))
    assert gen2.get_append_item_name(
        UDateTimePatternField.UDATPG_ERA_FIELD) == "epoch"

    gen2.set_append_item_name(
        UDateTimePatternField.UDATPG_ERA_FIELD,
        "era")
    assert gen2.get_append_item_name(
        UDateTimePatternField.UDATPG_ERA_FIELD) == "era"

    # UnicodeString icu::DateTimePatternGenerator::getBaseSkeleton(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    result = gen1.get_base_skeleton(UnicodeString("HH:mm"))
    assert isinstance(result, UnicodeString)
    assert result == "Hm"

    result = gen1.get_base_skeleton("HH:mm")
    assert isinstance(result, UnicodeString)
    assert result == "Hm"

    # StringEnumeration *icu::DateTimePatternGenerator::getBaseSkeletons(
    #       UErrorCode &status
    # )
    it = gen1.get_base_skeletons()
    assert isinstance(it, StringEnumeration)
    assert len(it) > 0

    # [1]
    # UnicodeString icu::DateTimePatternGenerator::getBestPattern(
    #       const UnicodeString &skeleton,
    #       UDateTimePatternMatchOptions options,
    #       UErrorCode &status
    # )
    result = gen1.get_best_pattern(
        UnicodeString("MMMMdjmm"),
        UDateTimePatternMatchOptions.UDATPG_MATCH_NO_OPTIONS)
    assert isinstance(result, UnicodeString)
    assert result == "MMMM d, h:mm a"

    result = gen1.get_best_pattern(
        "MMMMdjmm",
        UDateTimePatternMatchOptions.UDATPG_MATCH_NO_OPTIONS)
    assert isinstance(result, UnicodeString)
    assert result == "MMMM d, h:mm a"

    # [2]
    # UnicodeString icu::DateTimePatternGenerator::getBestPattern(
    #       const UnicodeString &skeleton,
    #       UErrorCode &status
    # )
    result = gen1.get_best_pattern(UnicodeString("MMMMdjmm"))
    assert isinstance(result, UnicodeString)
    assert result == "MMMM d, h:mm a"

    result = gen1.get_best_pattern("MMMMdjmm")
    assert isinstance(result, UnicodeString)
    assert result == "MMMM d, h:mm a"

    # const UnicodeString &icu::DateTimePatternGenerator::getDateTimeFormat()
    result = gen2.get_date_time_format()
    assert isinstance(result, UnicodeString)
    assert result == "{1}, {0}"

    # void icu::DateTimePatternGenerator::setDateTimeFormat(
    #       const UnicodeString &dateTimeFormat
    # )
    gen2.set_date_time_format(UnicodeString("{1} 'at' {0}"))
    assert gen2.get_date_time_format() == "{1} 'at' {0}"

    gen2.set_date_time_format("{1}, {0}")
    assert gen2.get_date_time_format() == "{1}, {0}"

    # const UnicodeString &icu::DateTimePatternGenerator::getDecimal()
    result = gen2.get_decimal()
    assert isinstance(result, UnicodeString)
    assert result == "."

    # void icu::DateTimePatternGenerator::setDecimal(
    #       const UnicodeString &decimal
    # )
    gen2.set_decimal(UnicodeString(","))
    assert gen2.get_decimal() == ","

    gen2.set_decimal(".")
    assert gen2.get_decimal() == "."

    # const UnicodeString &
    # icu::DateTimePatternGenerator::getPatternForSkeleton(
    #       const UnicodeString &skeleton
    # )
    result = gen1.get_pattern_for_skeleton(UnicodeString("yMMMd"))
    assert isinstance(result, UnicodeString)
    assert result == "MMM d, y"

    result = gen1.get_pattern_for_skeleton("yMMMd")
    assert isinstance(result, UnicodeString)
    assert result == "MMM d, y"

    # UnicodeString icu::DateTimePatternGenerator::getSkeleton(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    result = gen1.get_skeleton(UnicodeString("MMM d, y"))
    assert isinstance(result, UnicodeString)
    assert result == "yMMMd"

    result = gen1.get_skeleton("MMM d, y")
    assert isinstance(result, UnicodeString)
    assert result == "yMMMd"

    # StringEnumeration *icu::DateTimePatternGenerator::getSkeletons(
    #       UErrorCode &status
    # )
    it = gen1.get_skeletons()
    assert isinstance(it, StringEnumeration)
    assert len(it) > 0

    # [1]
    # UnicodeString icu::DateTimePatternGenerator::replaceFieldTypes(
    #       const UnicodeString &pattern,
    #       const UnicodeString &skeleton,
    #       UDateTimePatternMatchOptions options,
    #       UErrorCode &status
    # )
    result = gen1.replace_field_types(
        UnicodeString("d-M H:m"),
        UnicodeString("MMMMddHHmm"),
        UDateTimePatternMatchOptions.UDATPG_MATCH_ALL_FIELDS_LENGTH)
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM HH:mm"

    result = gen1.replace_field_types(
        "d-M H:m",
        UnicodeString("MMMMddHHmm"),
        UDateTimePatternMatchOptions.UDATPG_MATCH_ALL_FIELDS_LENGTH)
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM HH:mm"

    result = gen1.replace_field_types(
        UnicodeString("d-M H:m"),
        "MMMMddHHmm",
        UDateTimePatternMatchOptions.UDATPG_MATCH_ALL_FIELDS_LENGTH)
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM HH:mm"

    result = gen1.replace_field_types(
        "d-M H:m",
        "MMMMddHHmm",
        UDateTimePatternMatchOptions.UDATPG_MATCH_ALL_FIELDS_LENGTH)
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM HH:mm"

    # [2]
    # UnicodeString icu::DateTimePatternGenerator::replaceFieldTypes(
    #       const UnicodeString &pattern,
    #       const UnicodeString &skeleton,
    #       UErrorCode &status
    # )
    result = gen1.replace_field_types(
        UnicodeString("d-M H:m"),
        UnicodeString("MMMMddHHmm"))
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM H:m"

    result = gen1.replace_field_types(
        "d-M H:m",
        UnicodeString("MMMMddHHmm"))
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM H:m"

    result = gen1.replace_field_types(
        UnicodeString("d-M H:m"),
        "MMMMddHHmm")
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM H:m"

    result = gen1.replace_field_types(
        "d-M H:m",
        "MMMMddHHmm")
    assert isinstance(result, UnicodeString)
    assert result == "dd-MMMM H:m"


def test_clone():
    gen1 = DateTimePatternGenerator.create_instance("en_US")

    # DateTimePatternGenerator *icu::DateTimePatternGenerator::clone()
    gen2 = gen1.clone()
    assert isinstance(gen2, DateTimePatternGenerator)
    assert gen2 == gen1

    gen3 = copy.copy(gen1)
    assert gen3 == gen1

    gen4 = copy.deepcopy(gen1)
    assert gen4 == gen1


def test_create_empty_instance():
    # static DateTimePatternGenerator *
    # icu::DateTimePatternGenerator::createEmptyInstance(UErrorCode &status)
    gen = DateTimePatternGenerator.create_empty_instance()
    assert isinstance(gen, DateTimePatternGenerator)

    # UDateTimePatternConflict icu::DateTimePatternGenerator::addPattern(
    #       const UnicodeString &pattern,
    #       UBool override,
    #       UnicodeString &conflictingPattern,
    #       UErrorCode &status
    # )
    conflicting_pattern = UnicodeString()
    result = gen.add_pattern(UnicodeString("MMMMd"),
                             True,
                             conflicting_pattern)
    assert result == UDateTimePatternConflict.UDATPG_NO_CONFLICT
    assert len(conflicting_pattern) == 0

    result = gen.add_pattern("HH:mm",
                             True,
                             conflicting_pattern)
    assert result == UDateTimePatternConflict.UDATPG_NO_CONFLICT
    assert len(conflicting_pattern) == 0

    result = gen.add_pattern("MMMMMdd",
                             True,
                             conflicting_pattern)
    assert result == UDateTimePatternConflict.UDATPG_NO_CONFLICT
    assert len(conflicting_pattern) == 0

    # StringEnumeration *icu::DateTimePatternGenerator::getBaseSkeletons(
    #       UErrorCode &status
    # )
    it1 = gen.get_base_skeletons()
    assert isinstance(it1, StringEnumeration)
    assert list(it1) == ["Hm", "MMMMd", "MMMMMd"]

    # StringEnumeration *icu::DateTimePatternGenerator::getSkeletons(
    #       UErrorCode &status
    # )
    it2 = gen.get_skeletons()
    assert isinstance(it2, StringEnumeration)
    assert list(it2) == ["HHmm", "MMMMd", "MMMMMdd"]


def test_create_instance():
    # [1]
    # static DateTimePatternGenerator *
    # icu::DateTimePatternGenerator::createInstance(
    #       const Locale &uLocale,
    #       UErrorCode &status
    # )
    gen1 = DateTimePatternGenerator.create_instance(Locale("en_US"))
    assert isinstance(gen1, DateTimePatternGenerator)

    gen1a = DateTimePatternGenerator.create_instance("en_US")
    assert isinstance(gen1a, DateTimePatternGenerator)

    # [2]
    # static DateTimePatternGenerator *
    # icu::DateTimePatternGenerator::createInstance(UErrorCode &status)
    gen2 = DateTimePatternGenerator.create_instance()
    assert isinstance(gen2, DateTimePatternGenerator)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 67, reason="ICU4C<67")
def test_get_default_hour_cycle():
    from icupy import UDateFormatHourCycle

    gen = DateTimePatternGenerator.create_instance(Locale("en_US"))

    # UDateFormatHourCycle icu::DateTimePatternGenerator::getDefaultHourCycle(
    #       UErrorCode &status
    # )
    assert (gen.get_default_hour_cycle()
            == UDateFormatHourCycle.UDAT_HOUR_CYCLE_12)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 61, reason="ICU4C<61")
def test_get_field_display_name():
    from icupy import UDateTimePGDisplayWidth

    gen = DateTimePatternGenerator.create_instance(Locale("en_US"))

    # UnicodeString icu::DateTimePatternGenerator::getFieldDisplayName(
    #       UDateTimePatternField field,
    #       UDateTimePGDisplayWidth width
    # )
    result = gen.get_field_display_name(
        UDateTimePatternField.UDATPG_ERA_FIELD,
        UDateTimePGDisplayWidth.UDATPG_ABBREVIATED)
    assert isinstance(result, UnicodeString)
    assert result == "era"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 56, reason="ICU4C<56")
def test_static_get_skeleton():
    # static UnicodeString
    # icu::DateTimePatternGenerator::staticGetBaseSkeleton(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    result = DateTimePatternGenerator.static_get_base_skeleton(
        UnicodeString("d-M H:m"))
    assert isinstance(result, UnicodeString)
    assert result == "MdHm"

    result = DateTimePatternGenerator.static_get_base_skeleton(
        "d-M H:m")
    assert isinstance(result, UnicodeString)
    assert result == "MdHm"

    # static UnicodeString icu::DateTimePatternGenerator::staticGetSkeleton(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    result = DateTimePatternGenerator.static_get_skeleton(
        UnicodeString("d-M H:m"))
    assert isinstance(result, UnicodeString)
    assert result == "MdHm"

    result = DateTimePatternGenerator.static_get_skeleton(
        "d-M H:m")
    assert isinstance(result, UnicodeString)
    assert result == "MdHm"
